"""
NRV-axons
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler
import os
import time
import traceback
from abc import abstractmethod

import neuron
import numpy as np

from ..backend.file_handler import json_dump
from ..backend.log_interface import rise_error, rise_warning
from ..backend.NRV_Simulable import NRV_simulable
from ..fmod.electrodes import *
from ..fmod.extracellular import *
from ..fmod.recording import *

# Handling verbosity
Verbose = True

# enable faulthandler to ease "segmentation faults" debug
faulthandler.enable()

# instructions for Neuron, find mod files and hide GUI
dir_path = os.environ["NRVPATH"] + "/_misc"
neuron.load_mechanisms(dir_path + "/mods")
neuron.h.load_file("stdrun.hoc")

# list of supported models
unmyelinated_models = [
    "HH",
    "Rattay_Aberham",
    "Sundt",
    "Tigerholm",
    "Schild_94",
    "Schild_97",
]
myelinated_models = [
    "MRG",
    "Gaines_motor",
    "Gaines_sensory",
]
thin_myelinated_models = [
    "Extended_Gaines",
    "RGK",
]


##############################
## Usefull Neuron functions ##
##############################
def purge_neuron():
    """
    This function clears all the sections declared in neuron.h
    """
    for section in neuron.h.allsec():
        section = None


def d_lambda_rule(L, d_lambda, f, sec):
    """
    computes the d_lambda rule for a sections and returns the number of segments.

    Attributes
    ----------
    L           : int
        length of the section (um)
    d_lambda    : float
        d_lambda value, usually between 0.1 and 0.3
    f           : float
        maximal ionic current frequency (Hz)
    sec         : neuron section
        current neuron section on which to compute dlambda rule

    Returns
    -------
    Nseg        : int
        number of segment recommended to the section
    """
    return int((L / (d_lambda * neuron.h.lambda_f(f, sec=sec)) + 0.999) / 2) * 2 + 1


def create_Nseg_freq_shape(N_sec, shape, freq, freq_min, alpha_max):
    """
    creates a vector with the size of N_sec, a number of section to create an axon,
    and creates a shape of frequency to use with the d-lambda rule along the axon.

    Attributes
    ----------
    Nsec        : int
        the number of sections that will be used to implement the axon
    shape       : str
        shape of the output vector, pick between:
            "pyramidal"         -> min frequencies on both sides and linear increase up to the middle at the maximum frequency
            "sigmoid"           -> same a befor with sigmoid increase instead of linear
            "plateau"           -> sale as pyramidal except the max frequency is holded on a central plateau
            "plateau_sigmoid"   -> same as previous with sigmoid increase
    freq        : float
        maximum value of the frequency for the d-lambda rule
    freq_min    : float
        minimum value of the frequency for the d-lambda rule
    alpha_max   : float
        proportion of the axon at the maximum frequency for plateau shapes

    Returns
    -------
    freqs   : np.array
        vector of frequency to apply to a multi-sections axon
    """
    if shape == "pyramidal":
        if (N_sec % 2) == 0:
            X = np.concatenate(
                [
                    np.linspace(0, 1, num=int(N_sec / 2)),
                    np.linspace(1, 0, num=int(N_sec / 2)),
                ]
            )
        else:
            X = np.concatenate(
                [
                    np.linspace(0, 1, num=int(N_sec / 2)),
                    np.asarray([1]),
                    np.linspace(1, 0, num=int(N_sec / 2)),
                ]
            )
    elif shape == "sigmoid":
        if (N_sec % 2) == 0:
            x_1 = np.linspace(-6, 6, num=int(N_sec / 2))
            x_2 = np.linspace(6, -6, num=int(N_sec / 2))
            X = np.concatenate(
                [
                    (1 / (1 + np.exp(-x_1))),
                    (1 / (1 + np.exp(-x_2))),
                ]
            )
        else:
            x_1 = np.linspace(-6, 6, num=int(N_sec / 2))
            x_2 = np.linspace(6, -6, num=int(N_sec / 2))
            X = np.concatenate(
                [
                    (1 / (1 + np.exp(-x_1))),
                    np.asarray([1]),
                    (1 / (1 + np.exp(-x_2))),
                ]
            )
    elif shape == "plateau_sigmoid":
        length = int(N_sec * (1.0 - alpha_max) / 2)
        x_1 = np.linspace(-6, 6, num=length)
        x_2 = np.linspace(6, -6, num=length)
        X = np.concatenate(
            [
                (1 / (1 + np.exp(-x_1))),
                np.ones(int(N_sec) - 2 * length),
                (1 / (1 + np.exp(-x_2))),
            ]
        )
    else:
        # pyramid with plateau shape
        length = int(N_sec * (1.0 - alpha_max) / 2)
        X = np.concatenate(
            [
                np.linspace(0, 1, num=length),
                np.ones(N_sec - 2 * length),
                np.linspace(1, 0, num=length),
            ]
        )
    # applying the shape to the number of sections between the two specified frequencies
    freqs = X * (freq - freq_min) + freq_min
    return freqs


def rotate_list(l, n):
    """
    rotate a list with a defined number of indexes to shift

    Attributes
    ----------
    l   : list
        list to rotate
    n   : int
        number of indexes to r

    Returns
    -------
    l   : list
    """
    return l[-n:] + l[:-n]


#############################
## Generic class for axons ##
#############################
class axon(NRV_simulable):
    """
    Axon is a generic object to describe an axonal fiber,
    (soma and interconnection are not taken into account, all axons are independant from others)
    From user perspective, call myelinated, unmylinated or thin_myelinated classes for robust code. All axons
    are supposed to be along the x-axis, as Neuron would do when creatnig a basic longitudinal geometry.

    WARNING: do not create more than one axon at a time for one process, to prevent from parameters overlaps in Neuron
    """

    @abstractmethod
    def __init__(
        self,
        y=0,
        z=0,
        d=1,
        L=100,
        dt=0.001,
        Nseg_per_sec=0,
        freq=100,
        freq_min=0,
        mesh_shape="plateau_sigmoid",
        alpha_max=0.3,
        d_lambda=0.1,
        v_init=None,
        T=None,
        ID=0,
        threshold=-40,
        **kwarks,
    ):
        """
        initialisation of the axon,

        Parameters
        ----------
        y               : float
            y-axis coordinate of the axon (um)
        z               : float
            z-axis coordinate of the axon (um)
        d               : float
            axon diameter (um)
        L               : float
            axon length (um)
        dt              : float
            simulation time stem for Neuron (ms), by default 1us
        Nseg_per_sec    : float
            number of segment per section in Neuron. If set to 0, the number of segment per section is calculated with the d-lambda rule
        freq            : float
            frequency of the d-lambda rule (Hz), by default 100Hz
        freq_min        : float
            minimum frequency for the d-lambda rule when meshing is irregular, 0 for regular meshing
        v_init          : float
            initial value for the membrane voltage (mV), specify None for automatic model choice of v_init
        T               : float
            temperature (C), specify None for automatic model choice of temperature
        ID              : int
            axon ID, by default set to 0
        threshold       : int
            membrane voltage threshold for spike detection (mV), by default -40mV
        """
        super().__init__()
        self.type = "axon"
        ## Given by user
        self.x = []
        self.y = y
        self.z = z
        self.d = d
        self.L = L
        self.dt = dt
        self.Nseg_per_sec = Nseg_per_sec
        self.freq = freq
        self.freq_min = freq_min
        self.mesh_shape = mesh_shape
        self.alpha_max = alpha_max
        self.d_lambda = d_lambda
        self.v_init = v_init
        self.T = T
        self.ID = ID
        self.threshold = threshold
        self.model = None
        self.t_sim = 0
        self.sim_time = 0
        self.timeVector = None
        self.t_len = 0
        ## internal use
        self.created = False
        self.Nsec = 0
        self.Nseg = 0
        self.myelinated = ""
        ## stims
        self.intra_current_stim = []
        self.intra_current_stim_positions = []
        self.intra_current_stim_starts = []
        self.intra_current_stim_durations = []
        self.intra_current_stim_amplitudes = []
        self.intra_voltage_stim = None
        self.intra_voltage_stim_position = []
        self.intra_voltage_stim_stimulus = None
        self.extra_stim = None
        self.footprints = None
        ## recording mechanism
        self.record = False
        self.recorder = None
        ## abstract parameters
        self.x_rec = np.asarray([])
        self.rec = "all"
        self.node_index = None
        ## required for generate_axon_from_results
        if "diameter" in kwarks:
            self.d = kwarks["diameter"]

    def __del__(self):
        for section in neuron.h.allsec():
            section = None
        super().__del__()

    def __clear_reclists(self):
        keys = list(self.__dict__.keys())
        for key in keys:
            if 'reclist' in key:
                del self.__dict__[key]

    def save(
        self,
        save=False,
        fname="axon.json",
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        blacklist=[],
    ):
        """
        Return axon as dictionary and eventually save it as json file
        WARNING to use BEFORE stimulation, (for post simulation savings use
        save_axon_results_as_json on results dictionary

        Parameters
        ----------
        save    : bool
            if True, save in json files
        fname   : str
            Path and Name of the saving file, by default "axon.json"

        Returns
        -------
        ax_dic : dict
            dictionary containing all information
        """
        bc = [i for i in blacklist]
        bc += ["extra_stim", "intra_current_stim", "intra_voltage_stim"]
        if not intracel_context:
            bc += [
                "intra_current_stim_positions",
                "intra_current_stim_starts",
                "intra_current_stim_durations",
                "intra_current_stim_amplitudes",
                "intra_voltage_stim_position",
                "intra_voltage_stim_stimulus",
                "intra_voltage_stim_stimulus",
                "intra_voltage_stim_stimulus",
            ]
        if not extracel_context:
            bc += [
                "footprints",
            ]
        if not rec_context:
            bc += [
                "record",
                "recorder",
            ]

        key_dict = super().save(blacklist=bc)

        # As to be done separatly because mph object cannot be deep copied
        if extracel_context:
            key_dict["extra_stim"] = self.extra_stim.save()

        if save:
            json_dump(key_dict, fname)
        return key_dict

    def load(
        self,
        data,
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        blacklist=[],
    ):
        """
        Load all axon properties from a dictionary or a json file

        Parameters
        ----------
        data    : str or dict
            json file path or dictionary containing axon information
        """
        bc = [i for i in blacklist]
        if not intracel_context:
            bc += [
                "intra_current_stim_positions",
                "intra_current_stim_starts",
                "intra_current_stim_durations",
                "intra_current_stim_amplitudes",
                "intra_voltage_stim_position",
                "intra_voltage_stim_stimulus",
                "intra_voltage_stim_stimulus",
                "intra_voltage_stim_stimulus",
            ]
        if not extracel_context:
            bc += ["footprints", "extra_stim"]
        if not rec_context:
            bc += [
                "record",
                "recorder",
            ]

        super().load(data=data, blacklist=bc)
        com = "self._" + self.nrv_type + "__compute_axon_parameters()"
        eval(com)
        if intracel_context:
            for i in range(len(self.intra_current_stim_positions)):
                position = self.intra_current_stim_positions[i]
                stim_start = self.intra_current_stim_starts[i]
                duration = self.intra_current_stim_durations[i]
                amplitude = self.intra_current_stim_amplitudes[i]
                self.insert_I_Clamp(position / self.L, stim_start, duration, amplitude)
            if self.intra_voltage_stim_stimulus is not None:
                position = self.intra_voltage_stim_position[0]
                self.insert_V_Clamp(position / self.L, self.intra_voltage_stim_stimulus)

    def save_axon(
        self,
        save=False,
        fname="axon.json",
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
    ):
        rise_warning("save_axon is a deprecated method use save")
        self.save(
            save=save,
            fname=fname,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
        )

    def load_axon(
        self,
        data,
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
    ):
        rise_warning("load_axon is a deprecated method use load")
        self.save(data, extracel_context, intracel_context, rec_context)

    def __define_shape(self):
        """
        Define the shape of the axon after all sections creation
        """
        neuron.h.define_shape()

    def topology(self):
        """
        call the neuron topology function to plot the current topology on prompt
        """
        neuron.h.topology()

    def __get_allseg_positions(self):
        """
        get the positions of all segments define by neuron using allsec()
        WARNING: for internal use only, this is used to get all  position for external field computation
        in extracellular stimulation methods. This methods ensures to take into accounts all segments in the
        simulation

        Returns
        -------
        x   : np.array
            array of x coordinates of the computation points, in um
        y   : float
            y coordinate of the axon, in um
        z   : float
            z coordinate of the axon, in um
        """
        y = self.y
        z = self.z
        x = []
        for sec in neuron.h.allsec():
            offset = sec.x3d(0)
            for seg in sec:
                x.append(offset + seg.x * sec.L)
        return np.asarray(x), y, z

    def __set_allseg_vext(self, vext):
        """
        set the external potential of all segments accordingly to a vector

        Parameters
        ----------
        vext    : list or array of values to give to all segments e_extracellular in all segment of all sections in the simulation, in mV
        """
        k = 0
        for sec in neuron.h.allsec():
            for seg in sec:
                seg.e_extracellular = vext[k]
                k += 1

    def attach_extracellular_stimulation(self, stim):
        """
        attach a extracellular context of simulation for an axon

        Parameters
        ----------
        stim     : stimulation object
            see Extracellular.stimulation help for more details
        """
        if is_extra_stim(stim):
            self.extra_stim = stim

    def change_stimulus_from_elecrode(self, ID_elec, stimulus):
        """
        Change the stimulus of the ID_elec electrods

        Parameters:
        ----------
            ID_elec  : int
                ID of the electrode which should be changed
            stimulus  : stimulus
                New stimulus to set
        """
        if self.extra_stim is not None:
            self.extra_stim.change_stimulus_from_elecrode(ID_elec, stimulus)
        else:
            rise_warning("Cannot be changed: no extrastim in the axon")

    def get_electrodes_footprints_on_axon(
        self, save_ftp_only=False, filename="electrodes_footprint.ftpt"
    ):
        """
        get electrodes footprints on each axon segment

        Parameters
        ----------
        save_ftp_only        :bool
            if true save_ftp_only result in a .ftpt file
        filename    :str
            saving file name and path

        Returns
        -------
        footprints   :dict
        Dictionnary composed of extracellular footprint array, the keys are int value
        of the corresponding electrode ID
        """
        footprints = {}
        x, y, z = self.__get_allseg_positions()
        # self.extra_stim.synchronise_stimuli()
        self.extra_stim.compute_electrodes_footprints(x, y, z, self.ID)
        for i in range(len(self.extra_stim.electrodes)):
            elec = self.extra_stim.electrodes[i]
            footprints[i] = elec.footprint
        if save_ftp_only:
            json_dump(footprints, filename)
        self.footprints = footprints
        self.extra_stim.clear_electrodes_footprints()
        return footprints

    def attach_extracellular_recorder(self, rec):
        """
        attach an extracellular recorder to the axon

        Parameters
        ----------
        rec     : recorder object
            see Recording.recorder help for more details
        """
        if is_recorder(rec):
            self.record = True
            self.recorder = rec

    def shut_recorder_down(self):
        """
        Shuts down the recorder locally
        """
        self.record = False

    def simulate(
        self,
        **kwargs,
    ):
        """
        Simulates the axon using neuron framework

        Parameters
        ----------
        t_sim               : float
            total simulation time (ms), by default 20 ms
        self.record_V_mem        : bool
            if true, the membrane voltage is recorded, set to True by default
                see unmyelinated/myelinated to see where recording occur
                results stored with the key "V_mem"
        self.record_I_mem        : bool
            if true, the membrane current is recorded, set to False by default
        self.record_I_ions       : bool
            if true, the ionic currents are recorded, set to False by default
        record_particules   : bool
            if true, the marticule states are recorded, set to False by default
        self.loaded_footprints           :dict or bool
        Dictionnary composed of extracellular footprint array, the keys are int value
        of the corresponding electrode ID, if None, footprints calculated during the simulation,
        set to None by default
        Returns
        -------
        axon_sim    : dictionnary
            all informations on neuron, segment position and all simulation results
        """
        axon_sim = super().simulate(**kwargs)
        axon_sim['diameter'] =self.d
        axon_sim['tstop'] =self.t_sim
        axon_sim.update(self.__add_extrastim_to_res())
        t_sim = self.t_sim
        # set recorders arrays - KEEP THIS CODE BEFORE INITIALISATION
        self.__set_time_recorder()
        if self.record_V_mem:
            self.set_membrane_voltage_recorders()
        if self.record_I_mem or self.record:
            self.set_membrane_current_recorders()
        if self.record_I_ions:
            self.set_ionic_current_recorders()
        if self.record_g_mem or self.record_g_ions:
            self.set_conductance_recorders()
        if self.record_particles:
            self.set_particules_values_recorders()
            if hasattr(self, "Markov_Nav_modeled_NoR"):
                self.set_Nav_recorders()

        ## initialisation and parameters for neuron - KEEP THIS CODE JUST BEFORE SIMULATION
        neuron.h.tstop = t_sim
        neuron.h.celsius = self.T  # set temperature in celsius
        neuron.h.finitialize(self.v_init)  # initialize voltage state
        neuron.h.v_init = self.v_init
        if self.dt == 0:
            rise_warning(
                "Warning: for Axon "
                + str(self.ID)
                + ", the ODE resolution is with adaptive step, atol set to 0.001 in Neuron"
            )
            neuron.h("cvode_active(1)")
            neuron.h("cvode.atol(0.001)")
        else:
            neuron.h.dt = self.dt  # set time step (ms)

        try:
            start_time = time.time()
            neuron.h.frecord_init()
            ###########################################
            #### THIS IS WHERE SIMULATION IS HANDLED ##
            ###########################################
            if self.extra_stim is not None:
                # prepare extracellular stimulation
                x, y, z = self.__get_allseg_positions()
                self.extra_stim.synchronise_stimuli()

                if self.loaded_footprints is None:
                    if self.footprints is None:
                        self.get_electrodes_footprints_on_axon()
                elif self.loaded_footprints:
                    if self.footprints is None:
                        rise_warning(
                            "no loaded footprints: computation will be done anyway"
                        )
                        self.get_electrodes_footprints_on_axon()
                elif not self.loaded_footprints:
                    self.get_electrodes_footprints_on_axon()
                self.extra_stim.set_electrodes_footprints(self.footprints)
                # compute the minimum time between stimuli changes, checks it"s not smaller than the computation dt, if so, there should be a warning to the user
                Delta_T_min = np.amin(np.diff(self.extra_stim.global_time_serie))
                if Delta_T_min < self.dt:
                    ## WARNING: the stimulus is over sampled compared to the neuron dt:
                    if Delta_T_min / self.dt < 1:  # HERE!! FOr dt test only
                        # if the stimulus minimal change time is more than 10% of user specified dt, change it to avoid problem
                        # prints a warning as computation time will increase !
                        new_dt = self.dt / np.ceil(self.dt / Delta_T_min)
                        axon_sim["dt"] = new_dt
                        rise_warning("Specified dt is too low")
                        rise_warning(
                            "... automatically changing the user specified dt from "
                            + str(self.dt)
                            + " to "
                            + str(new_dt)
                            + " ms"
                        )
                        self.dt = new_dt
                        neuron.h.dt = self.dt
                    else:
                        rise_warning(
                            "Neuron will undersample the stimulus... \
                            this may be unfortunate, consider to undersample your stimuli yourself,\
                            or chose a smaller dt"
                        )
                        rise_warning(
                            "... dt is "
                            + str(Delta_T_min / self.dt)
                            + " times bigger \
                            than the stimulus minimum time step difference"
                        )
                # loop on stimuli time serie
                for i in range(1, len(self.extra_stim.global_time_serie)):
                    t_step = min(self.extra_stim.global_time_serie[i], t_sim)
                    vext = self.extra_stim.compute_vext(i - 1)
                    self.__set_allseg_vext(vext)
                    neuron.h.continuerun(t_step)
                # finish simulation if needed
                if neuron.h.t < t_sim:
                    vext = self.extra_stim.compute_vext(
                        len(self.extra_stim.global_time_serie) - 1
                    )
                    self.__set_allseg_vext(vext)
                    neuron.h.continuerun(neuron.h.tstop)
            elif self.intra_voltage_stim is not None:
                # init if first point is at 0
                if self.intra_voltage_stim_stimulus.t[0] == 0:
                    self.intra_voltage_stim.amp[0] = self.intra_voltage_stim_stimulus.s[
                        0
                    ]
                # run simulation with a loop on the voltage clamp times
                for i in range(1, len(self.intra_voltage_stim_stimulus.t)):
                    t_step = min(self.intra_voltage_stim_stimulus.t[i], t_sim)
                    # run simulation
                    neuron.h.continuerun(t_step)
                    # apply new voltage clamp value
                    self.intra_voltage_stim.amp[0] = self.intra_voltage_stim_stimulus.s[
                        i
                    ]
                # finish simulation if needed
                if neuron.h.t < t_sim:
                    neuron.h.continuerun(neuron.h.tstop)
            else:
                neuron.h.continuerun(neuron.h.tstop)
            ###########################################
            ###########################################
            ###########################################
            self.sim_time = time.time() - start_time
            # simulation done, store results
            axon_sim["Simulation_state"] = "Successful"
            axon_sim["sim_time"] = self.sim_time
            axon_sim["t"] = self.__get_time_vector()
            axon_sim["x_rec"] = self.x_rec
            if self.myelinated and self.rec != "nodes":
                axon_sim["node_index"] = self.node_index
            if self.record_V_mem:
                axon_sim["V_mem"] = self.get_membrane_voltage()
            if self.record_I_mem or self.record:
                axon_sim["I_mem"] = self.get_membrane_current()
            if self.record_g_mem:
                axon_sim["g_mem"] = self.get_membrane_conductance()
            if self.record_g_ions:
                if not self.myelinated:
                    if self.model in ["HH", "Rattay_Aberham", "Sundt"]:
                        g_na_ax, g_k_ax, g_l_ax = self.get_ionic_conductance()
                        axon_sim["g_na"] = g_na_ax
                        axon_sim["g_k"] = g_k_ax
                        axon_sim["g_l"] = g_l_ax
                    elif self.model == "Tigerholm":
                        (
                            axon_sim["g_nav17"],
                            axon_sim["g_nav18"],
                            axon_sim["g_nav19"],
                            axon_sim["g_kA"],
                            axon_sim["g_kM"],
                            axon_sim["g_kdr"],
                            axon_sim["g_kna"],
                            axon_sim["g_h"],
                            axon_sim["g_naleak"],
                            axon_sim["g_kleak"],
                        ) = self.get_ionic_conductance()
                    else:
                        (
                            axon_sim["g_naf"],
                            axon_sim["g_nas"],
                            axon_sim["g_kd"],
                            axon_sim["g_ka"],
                            axon_sim["g_kds"],
                            axon_sim["g_kca"],
                            axon_sim["g_can"],
                            axon_sim["g_cat"],
                        ) = self.get_ionic_conductance()
                else:
                    if self.model == "MRG":
                        (
                            g_na_ax,
                            g_nap_ax,
                            g_k_ax,
                            g_l_ax,
                            g_i_ax,
                        ) = self.get_ionic_conductance()
                        axon_sim["g_na"] = g_na_ax
                        axon_sim["g_nap"] = g_nap_ax
                        axon_sim["g_k"] = g_k_ax
                        axon_sim["g_l"] = g_l_ax
                        axon_sim["g_i"] = g_i_ax
                    elif self.model in ["Gaines_motor", "Gaines_sensory"]:
                        (
                            g_na_ax,
                            g_nap_ax,
                            g_k_ax,
                            g_l_ax,
                            g_kf_ax,
                            g_q_ax,
                        ) = self.get_ionic_conductance()
                        axon_sim["g_na"] = g_na_ax
                        axon_sim["g_nap"] = g_nap_ax
                        axon_sim["g_k"] = g_k_ax
                        axon_sim["g_kf"] = g_kf_ax
                        axon_sim["g_l"] = g_l_ax
                        axon_sim["g_q"] = g_q_ax
                    else:
                        rise_warning(
                            "g_ions recorder not implemented for "
                            + self.model
                            + " model "
                        )
            if self.record_I_ions:
                if not self.myelinated:
                    if self.model in ["HH", "Rattay_Aberham", "Sundt"]:
                        I_na_ax, I_k_ax, I_l_ax = self.get_ionic_current()
                        axon_sim["I_na"] = I_na_ax
                        axon_sim["I_k"] = I_k_ax
                        axon_sim["I_l"] = I_l_ax
                    else:
                        I_na_ax, I_k_ax, I_ca_ax = self.get_ionic_current()
                        axon_sim["I_na"] = I_na_ax
                        axon_sim["I_k"] = I_k_ax
                        axon_sim["I_ca"] = I_ca_ax
                else:
                    if self.model == "MRG":
                        I_na_ax, I_nap_ax, I_k_ax, I_l_ax = self.get_ionic_current()
                        axon_sim["I_na"] = I_na_ax
                        axon_sim["I_nap"] = I_nap_ax
                        axon_sim["I_k"] = I_k_ax
                        axon_sim["I_l"] = I_l_ax
                    elif self.model in ["Gaines_motor", "Gaines_sensory"]:
                        (
                            I_na_ax,
                            I_nap_ax,
                            I_k_ax,
                            I_kf_ax,
                            I_q_ax,
                            I_l_ax,
                        ) = self.get_ionic_current()
                        axon_sim["I_na"] = I_na_ax
                        axon_sim["I_nap"] = I_nap_ax
                        axon_sim["I_k"] = I_k_ax
                        axon_sim["I_kf"] = I_kf_ax
                        axon_sim["I_q"] = I_q_ax
                        axon_sim["I_l"] = I_l_ax
                    elif self.model in ["Extended_Gaines"]:
                        (
                            I_na_ax,
                            I_nap_ax,
                            I_k_ax,
                            I_kf_ax,
                            I_l_ax,
                        ) = self.get_ionic_current()
                        axon_sim["I_na"] = I_na_ax
                        axon_sim["I_nap"] = I_nap_ax
                        axon_sim["I_k"] = I_k_ax
                        axon_sim["I_kf"] = I_kf_ax
                        axon_sim["I_l"] = I_l_ax
                    else:  # should be RGK
                        I_na_ax, I_k_ax, I_l_ax = self.get_ionic_current()
                        axon_sim["I_na"] = I_na_ax
                        axon_sim["I_k"] = I_k_ax
                        axon_sim["I_l"] = I_l_ax
            if self.record_particles:
                if not self.myelinated:
                    if self.model in ["HH", "Rattay_Aberham", "Sundt"]:
                        m_ax, n_ax, h_ax = self.get_particles_values()
                        axon_sim["m"] = m_ax
                        axon_sim["n"] = n_ax
                        axon_sim["h"] = h_ax
                    elif self.model in ["Tigerholm"]:
                        (
                            m_nav18_ax,
                            h_nav18_ax,
                            s_nav18_ax,
                            u_nav18_ax,
                            m_nav19_ax,
                            h_nav19_ax,
                            s_nav19_ax,
                            m_nattxs_ax,
                            h_nattxs_ax,
                            s_nattxs_ax,
                            n_kdr_ax,
                            m_kf_ax,
                            h_kf_ax,
                            ns_ks_ax,
                            nf_ks_ax,
                            w_kna_ax,
                            ns_h_ax,
                            nf_h_ax,
                        ) = self.get_particles_values()
                        axon_sim["m_nav18"] = m_nav18_ax
                        axon_sim["h_nav18"] = h_nav18_ax
                        axon_sim["s_nav18"] = s_nav18_ax
                        axon_sim["u_nav18"] = u_nav18_ax
                        axon_sim["m_nav19"] = m_nav19_ax
                        axon_sim["h_nav19"] = h_nav19_ax
                        axon_sim["s_nav19"] = s_nav19_ax
                        axon_sim["m_nattxs"] = m_nattxs_ax
                        axon_sim["h_nattxs"] = h_nattxs_ax
                        axon_sim["s_nattxs"] = s_nattxs_ax
                        axon_sim["n_kdr"] = n_kdr_ax
                        axon_sim["m_kf"] = m_kf_ax
                        axon_sim["h_kf"] = h_kf_ax
                        axon_sim["ns_ks"] = ns_ks_ax
                        axon_sim["nf_ks"] = nf_ks_ax
                        axon_sim["w_kna"] = w_kna_ax
                        axon_sim["ns_h"] = ns_h_ax
                        axon_sim["nf_h"] = nf_h_ax
                    elif self.model in ["Schild_94"]:
                        (
                            d_can_ax,
                            f1_can_ax,
                            f2_can_ax,
                            d_cat_ax,
                            f_cat_ax,
                            p_ka_ax,
                            q_ka_ax,
                            c_kca_ax,
                            n_kd_ax,
                            x_kds_ax,
                            y1_kds_ax,
                            m_naf_ax,
                            h_naf_ax,
                            l_naf_ax,
                            m_nas_ax,
                            h_nas_ax,
                        ) = self.get_particles_values()
                        axon_sim["d_can"] = d_can_ax
                        axon_sim["f1_can"] = f1_can_ax
                        axon_sim["f2_can"] = f2_can_ax
                        axon_sim["d_cat"] = d_cat_ax
                        axon_sim["f_cat"] = f_cat_ax
                        axon_sim["p_ka"] = p_ka_ax
                        axon_sim["q_ka"] = q_ka_ax
                        axon_sim["c_ka"] = c_kca_ax
                        axon_sim["n_kd"] = n_kd_ax
                        axon_sim["x_kds"] = x_kds_ax
                        axon_sim["y1_kds"] = y1_kds_ax
                        axon_sim["m_naf"] = m_naf_ax
                        axon_sim["h_naf"] = h_naf_ax
                        axon_sim["l_naf"] = l_naf_ax
                        axon_sim["m_nas"] = m_nas_ax
                        axon_sim["h_nas"] = h_nas_ax
                    else:  # should be "Schild_97"
                        (
                            d_can_ax,
                            f1_can_ax,
                            f2_can_ax,
                            d_cat_ax,
                            f_cat_ax,
                            p_ka_ax,
                            q_ka_ax,
                            c_kca_ax,
                            n_kd_ax,
                            x_kds_ax,
                            y1_kds_ax,
                            m_naf_ax,
                            h_naf_ax,
                            m_nas_ax,
                            h_nas_ax,
                        ) = self.get_particles_values()
                        axon_sim["d_can"] = d_can_ax
                        axon_sim["f1_can"] = f1_can_ax
                        axon_sim["f2_can"] = f2_can_ax
                        axon_sim["d_cat"] = d_cat_ax
                        axon_sim["f_cat"] = f_cat_ax
                        axon_sim["p_ka"] = p_ka_ax
                        axon_sim["q_ka"] = q_ka_ax
                        axon_sim["c_ka"] = c_kca_ax
                        axon_sim["n_kd"] = n_kd_ax
                        axon_sim["x_kds"] = x_kds_ax
                        axon_sim["y1_kds"] = y1_kds_ax
                        axon_sim["m_naf"] = m_naf_ax
                        axon_sim["h_naf"] = h_naf_ax
                        axon_sim["m_nas"] = m_nas_ax
                        axon_sim["h_nas"] = h_nas_ax
                else:
                    if self.model == "MRG":
                        m_ax, mp_ax, h_ax, s_ax = self.get_particles_values()
                        axon_sim["m"] = m_ax
                        axon_sim["mp"] = mp_ax
                        axon_sim["h"] = h_ax
                        axon_sim["s"] = s_ax
                    elif self.model in ["Gaines_motor", "Gaines_sensory"]:
                        (
                            m_ax,
                            mp_ax,
                            h_ax,
                            s_ax,
                            n_ax,
                            q_ax,
                        ) = self.get_particles_values()
                        axon_sim["m"] = m_ax
                        axon_sim["mp"] = mp_ax
                        axon_sim["h"] = h_ax
                        axon_sim["s"] = s_ax
                        axon_sim["n"] = n_ax
                        axon_sim["q"] = q_ax
                    elif self.model in ["Extended_Gaines"]:
                        m_ax, mp_ax, s_ax, h_ax, n_ax = self.get_particles_values()
                        axon_sim["m"] = m_ax
                        axon_sim["mp"] = mp_ax
                        axon_sim["s"] = s_ax
                        axon_sim["h"] = h_ax
                        axon_sim["n"] = n_ax
                    else:  # Should be RGK
                        (
                            RGK_m_nav1p9_ax,
                            RGK_h_nav1p9_ax,
                            RGK_s_nav1p9_ax,
                            RGK_m_nax_ax,
                            RGK_h_nax_ax,
                            RGK_ns_ks_ax,
                            RGK_nf_ks_ax,
                        ) = self.get_particles_values()
                        axon_sim["m_nav19"] = RGK_m_nav1p9_ax
                        axon_sim["h_nav19"] = RGK_h_nav1p9_ax
                        axon_sim["s_nav19"] = RGK_s_nav1p9_ax
                        axon_sim["m_nax"] = RGK_m_nax_ax
                        axon_sim["h_nax"] = RGK_h_nax_ax
                        axon_sim["ns_ks"] = RGK_ns_ks_ax
                        axon_sim["nf_ks"] = RGK_nf_ks_ax
                if hasattr(self, "Markov_Nav_modeled_NoR"):
                    (
                        I_nav11_ax,
                        C1_nav11_ax,
                        C2_nav11_ax,
                        O1_nav11_ax,
                        O2_nav11_ax,
                        I1_nav11_ax,
                        I2_nav11_ax,
                        I_nav16_ax,
                        C1_nav16_ax,
                        C2_nav16_ax,
                        O1_nav16_ax,
                        O2_nav16_ax,
                        I1_nav16_ax,
                        I2_nav16_ax,
                    ) = self.get_Nav_values()
                    axon_sim["I_Nav11"] = I_nav11_ax
                    axon_sim["C1_nav11"] = C1_nav11_ax
                    axon_sim["C2_nav11"] = C2_nav11_ax
                    axon_sim["O1_nav11"] = O1_nav11_ax
                    axon_sim["O2_nav11"] = O2_nav11_ax
                    axon_sim["I1_nav11"] = I1_nav11_ax
                    axon_sim["I2_nav11"] = I2_nav11_ax
                    axon_sim["I_Nav16"] = I_nav16_ax
                    axon_sim["C1_nav16"] = C1_nav16_ax
                    axon_sim["C2_nav16"] = C2_nav16_ax
                    axon_sim["O1_nav16"] = O1_nav16_ax
                    axon_sim["O2_nav16"] = O2_nav16_ax
                    axon_sim["I1_nav16"] = I1_nav16_ax
                    axon_sim["I2_nav16"] = I2_nav16_ax
            # check for extracellular potential Recording, initialize, compute footprint and get potential
            if self.record:
                # init the potential, if already done by another axon nothing should be performed
                self.recorder.init_recordings(len(axon_sim["t"]))
                # compute footprints
                if self.myelinated == True:
                    self.recorder.compute_footprints(
                        axon_sim["x_nodes"],
                        self.y,
                        self.z,
                        self.d,
                        self.ID,
                        self.myelinated,
                    )
                    # compute extra-cellular potential and add it to already computed ones
                else:
                    self.recorder.compute_footprints(
                        axon_sim["x_rec"],
                        self.y,
                        self.z,
                        self.d,
                        self.ID,
                        self.myelinated,
                    )
                    # compute extra-cellular potential and add it to already computed ones
                self.recorder.set_time(axon_sim["t"])
                self.recorder.add_axon_contribution(axon_sim["I_mem"], self.ID)

        except KeyboardInterrupt:
            """rise_error(
                "\n Caught KeyboardInterrupt, simulation stoped by user, stopping process..."
            )"""
            raise KeyboardInterrupt
        except:
            axon_sim["Simulation_state"] = "Unsuccessful"
            axon_sim["Error_from_prompt"] = traceback.format_exc()
            rise_warning(
                "Simulation induced an error while using neuron: \n"
                + axon_sim["Error_from_prompt"]
            )
            axon_sim["Neuron_t_max"] = neuron.h.t
        self.__clear_reclists()
        return axon_sim

    ##############################
    ## Result recording methods ##
    ##############################
    def __set_time_recorder(self):
        """
        internal use: set recorders in neuron for the time vector
        """
        self.timeVector = neuron.h.Vector().record(neuron.h._ref_t)

    def __get_time_vector(self):
        """
        internal use: get the time vector and stor its length for further use

        Returns
        -------
        t   : np.array
            time vector of the previous simulation, numpy array
        """
        t = np.array(self.timeVector)
        self.t_len = len(t)
        return t

    ##############################
    ## Result handleling method ##
    ##############################
    def __add_extrastim_to_res(self):
        """
        
        """
        axon_sim = {}
        axon_sim["intra_stim_positions"] = self.intra_current_stim_positions
        axon_sim["intra_stim_starts"] = self.intra_current_stim_starts
        axon_sim["intra_stim_durations"] = self.intra_current_stim_durations
        axon_sim["intra_stim_amplitudes"] = self.intra_current_stim_amplitudes
        if self.extra_stim != None:
            if is_analytical_extra_stim(self.extra_stim):
                # save material
                axon_sim["extracellular_context"] = "analytical"
                axon_sim["extracellular_material"] = self.extra_stim.material.name
                # save electrode positions
                electrodes_x = []
                electrodes_y = []
                electrodes_z = []
                for elec in self.extra_stim.electrodes:
                    electrodes_x.append(elec.x)
                    electrodes_y.append(elec.y)
                    electrodes_z.append(elec.z)
                axon_sim["extracellular_electrode_x"] = electrodes_x
                axon_sim["extracellular_electrode_y"] = electrodes_y
                axon_sim["extracellular_electrode_z"] = electrodes_z
                # save stimuli
                stimuli_list = []
                stimuli_time_list = []
                for stim in self.extra_stim.stimuli:
                    stimuli_list.append(stim.s)
                    stimuli_time_list.append(stim.t)
                axon_sim["extracellular_stimuli"] = stimuli_list
                axon_sim["extracellular_stimuli_t"] = stimuli_time_list
            else:
                axon_sim["extracellular_context"] = "FEM"
                axon_sim["model_name"] = self.extra_stim.model_fname
                axon_sim["endoneurium_material"] = self.extra_stim.endoneurium.name
                axon_sim["perineurium_material"] = self.extra_stim.perineurium.name
                axon_sim["epineurium_material"] = self.extra_stim.epineurium.name
                axon_sim["external_material"] = self.extra_stim.external_material.name
                # save electrode positions
                electrodes_x = []
                electrodes_type = []
                electrodes_y = []
                electrodes_z = []
                for electrode in self.extra_stim.electrodes:
                    if is_CUFF_electrode(electrode):
                        electrodes_x.append(electrode.x_center)
                        electrodes_y.append(0)
                        electrodes_z.append(0)
                        electrodes_type.append("CUFF")
                    else:
                        if is_LIFE_electrode(electrode):
                            electrodes_x.append(electrode.x + electrode.length / 2)
                            electrodes_type.append("LIFE")
                        electrodes_y.append(electrode.y)
                        electrodes_z.append(electrode.z)
                axon_sim["extracellular_electrode_type"] = electrodes_type
                axon_sim["extracellular_electrode_x"] = electrodes_x
                axon_sim["extracellular_electrode_y"] = electrodes_y
                axon_sim["extracellular_electrode_z"] = electrodes_z
                # save stimuli
                stimuli_list = []
                stimuli_time_list = []
                for stimulus in self.extra_stim.stimuli:
                    stimuli_list.append(stimulus.s)
                    stimuli_time_list.append(stimulus.t)
                axon_sim["extracellular_stimuli"] = stimuli_list
                axon_sim["extracellular_stimuli_t"] = stimuli_time_list
        return axon_sim

    ###########################
    ## Axon abstract methods ##
    ###########################
    @abstractmethod
    def insert_I_Clamp(self):
        pass

    @abstractmethod
    def insert_V_Clamp(self):
        pass

    @abstractmethod
    def set_membrane_voltage_recorders(self):
        pass

    @abstractmethod
    def set_membrane_current_recorders(self):
        pass

    @abstractmethod
    def set_ionic_current_recorders(self):
        pass

    @abstractmethod
    def set_conductance_recorders(self):
        pass

    @abstractmethod
    def set_particules_values_recorders(self):
        pass

    @abstractmethod
    def get_membrane_voltage(self):
        pass

    @abstractmethod
    def get_membrane_current(self):
        pass

    @abstractmethod
    def get_ionic_current(self):
        pass

    def get_membrane_conductance(self):
        pass

    def get_membrane_capacitance(self):
        pass

    def get_particules_values(self):
        pass


class axon_test(axon):
    """
    TO DO
    """

    def __init__(
        self,
        y,
        z,
        d,
        L,
        dt=0.001,
        Nseg_per_sec=0,
        freq=100,
        freq_min=0,
        mesh_shape="plateau_sigmoid",
        alpha_max=0.3,
        d_lambda=0.1,
        v_init=None,
        T=None,
        ID=0,
        threshold=-40,
    ):
        super().__init__(
            y,
            z,
            d,
            L,
            dt=0.001,
            Nseg_per_sec=0,
            freq=100,
            freq_min=0,
            mesh_shape="plateau_sigmoid",
            alpha_max=0.3,
            d_lambda=0.1,
            v_init=None,
            T=None,
            ID=0,
            threshold=-40,
        )

    def insert_I_Clamp(self):
        pass

    def insert_V_Clamp(self):
        pass

    def set_membrane_voltage_recorders(self):
        pass

    def set_membrane_current_recorders(self):
        pass

    def set_ionic_current_recorders(self):
        pass

    def set_conductance_recorders(self):
        pass

    def set_particules_values_recorders(self):
        pass

    def get_membrane_voltage(self):
        pass

    def get_membrane_current(self):
        pass

    def get_ionic_current(self):
        pass



"""
NRV-Cellular Level simulations
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
from ..backend.MCore import *
from ..fmod.electrodes import *
from ..fmod.extracellular import *
from ..fmod.materials import *
from ..fmod.stimulus import *
from ..nmod.axons import *
from ..nmod.fascicles import *
from ..nmod.myelinated import *
from ..nmod.unmyelinated import *

###############################################################
#########################  Loaders  ###########################
###############################################################


def load_any_fascicle(
    data, extracel_context=False, intracel_context=False, rec_context=False
):
    """
    generate any kind of fascicle from a dictionary or a json file

    Parameters
    ----------
    data    : str or dict
        json file path or dictionary containing fascicle information
    """
    synchronize_processes()
    if type(data) == str:
        fasc_dic = json_load(data)
    else:
        fasc_dic = data
    fasc = fascicle()
    fasc.load(
        fasc_dic,
        extracel_context=extracel_context,
        intracel_context=intracel_context,
        rec_context=rec_context,
    )
    if extracel_context and rec_context:
        return fasc, fasc.extra_stim, fasc.recorder
    elif extracel_context:
        return fasc, fasc.extra_stim
    elif rec_context:
        return fasc, fasc.recorder
    else:
        return fasc


def load_any_axon(
    data, extracel_context=False, intracel_context=False, rec_context=False
):
    """
    generate any kind of axon from a dictionary or a json file

    Parameters
    ----------
    data    : str or dict
        json file path or dictionary containing axon information
    """
    if type(data) == str:
        ax_dic = json_load(data)
    else:
        ax_dic = data

    if ax_dic["myelinated"] is True:
        ax = myelinated(0, 0, 1, 10)
    else:
        ax = unmyelinated(0, 0, 1, 10)

    ax.load(
        ax_dic,
        extracel_context=extracel_context,
        intracel_context=intracel_context,
        rec_context=rec_context,
    )
    if extracel_context and rec_context:
        return ax, ax.extra_stim, ax.recorder
    elif extracel_context:
        return ax, ax.extra_stim
    elif rec_context:
        return ax, ax.recorder
    else:
        return ax



"""
NRV-Cellular Level postprocessing
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler
import os

import matplotlib.pyplot as plt

from ...backend.file_handler import json_load
from ...backend.log_interface import pass_info
from ...fmod.electrodes import is_FEM_electrode
from ..cell.CL_postprocessing import *

# enable faulthandler to ease 'segmentation faults' debug
faulthandler.enable()


#############################
####### File handeler #######
#############################


def ls_axons_results(dir_path):
    """
    return list of axons simulation result files

    Parameters
    ----------
    dir_path     : str
        path the results directory

    Returns
    -------
    files        :list of str
        list of axons result files
    """
    list_files = [file for file in os.listdir(dir_path) if file[0:9] == "sim_axon_"]
    return list_files

def ls_csv(dir_path):
    """
    return list of axons simulation result files

    Parameters
    ----------
    dir_path     : str
        path the results directory

    Returns
    -------
    files        :list of str
        list of axons result files
    """
    list_files = [file for file in os.listdir(dir_path) if file[-4:] == ".csv"]
    return list_files


def rm_file(file_path, verbose=True):
    """
    Delete file

    Parameters
    ----------
    file_path : str
        path and name of the file to remove
    verbose     : str
        pass information when file is deleted
    """
    os.remove(file_path)
    pass_info("folowing file removed :" + file_path, verbose=verbose)


def rm_sim_dir(dir_path, verbose=True):
    """
    Delete directory
    Warning: use with caution deleted files cannot be recovered

    Parameters
    ----------
    file_path : str
        path and name of the file to remove
    verbose     : str
        pass information when file is deleted
    """
    if os.path.exists(dir_path):
        # messaging saying folder not empty
        if os.path.exists(
            dir_path + "00_Fascicle_config.json",
        ):
            rm_file(dir_path + "00_Fascicle_config.json", verbose)

        for file in ls_axons_results(dir_path)+ls_csv(dir_path):
            rm_file(dir_path + file, verbose)
        # checking whether the folder is empty or not
        if len(os.listdir(dir_path)) == 0:
            os.rmdir(dir_path)
            pass_info("folowing folder removed :" + dir_path, verbose=verbose)
        else:
            os.rmdir(dir_path)
            pass_info(
                "Folder contains files or folders which cannot be deleted",
                verbose=verbose,
            )
    else:
        # file not found message
        pass_info("Folder not found in the directory", verbose=verbose)

def rm_sim_dir_from_results(results, verbose=True):
    """
    Delete directory
    Warning: use with caution deleted files cannot be recovered

    Parameters
    ----------
    file_path : str
        path and name of the file to remove
    verbose     : str
        pass information when file is deleted
    """
    fasc_dir = results["save_path"] + "Fascicle_" +str(results["ID"]) + "/"
    rm_sim_dir(dir_path=fasc_dir, verbose=verbose)

def CAP_time_detection(Voltage, t, t_stim=0, stim_duration=0,tol=0.05, myelinated=False,\
    index=True):
    """
    internal use, Return index in the time scale or time of the start and stop of
    a Compound Action Potentiel 

    Parameters
    ----------
    Voltage : list(float)
        list of voltage value
    t     : list(float)
        list of time value
    t_stim : float
        time of the stimulation (to skip the stimulation artefact)
    stim_duration     : list(float)
        list of time value
    tol : float
        absolute tolerence in Volt
    index     : bool
        if true the time index is returned, else the time value
    """
    i_start_unm, i_stop_unm = 0, 0
    i_start_m, i_stop_m = 0, 0
    dt = t[1] - t[0]
    offset = int((t_stim + stim_duration)/dt)+1

    S = Voltage
    N = len(S) - offset

    for i in range(N-1):
        if i_start_unm == 0 and S[i+offset]- S[-1] > tol:
            i_start_unm = i+offset
        if i_stop_unm == 0 and abs(S[len(S)-i-1])- abs(S[-1]) > tol:
            i_stop_unm = len(S) - i - 1

    if myelinated is False: 
        i_max = i_start_unm + np.argmax(Voltage[i_start_unm:i_stop_unm])
        i_min = i_start_unm + np.argmin(Voltage[i_start_unm:i_stop_unm])
        if index:
            return i_start_unm, i_max, i_min, i_stop_unm
        else:
            return t[i_start_unm], t[i_max], t[i_min], t[i_stop_unm]
    else:
        ## No unmyelinated CAP found
        if i_start_unm == 0:
            i_stop_m = i_stop_unm
            for i in range(N-1):
                if i_start_m == 0 and abs(S[i+offset])-abs(S[-1])> tol:
                    i_start_m = i+offset
        else:
            N = i_start_unm - offset
            for i in range(N-1):
                if i_start_m == 0 and abs(S[i+offset])- abs(S[-1]) > tol:
                    i_start_m = i+offset
                if i_stop_m == 0 and abs(S[i_start_unm-i-1])- abs(S[-1]) > tol:
                    i_stop_m = i_start_unm - i - 1
        i_max = np.argmax(Voltage[offset:])
        i_min = i_start_m + np.argmin(Voltage[i_start_m:i_stop_m])

        if index:
            return i_start_m, i_max, i_min, i_stop_m
        else:
            return t[i_start_m], t[i_max], t[i_min], t[i_stop_m]

#############################
##### Result processing #####
#############################
def fascicular_state(
    dir_path,
    save=False,
    saving_file="facsicular_state.json",
    delete_files=False,
    verbose=True,
):
    """
    Return each axon caracteristics (blocked, Onset response, ...)

    Parameters
    ----------
    dir_path     : str
        path the results directory
    save        : bool
        if True save result in json file
    saving_file : str
        if save is True path and name of the saving file

    Returns
    -------
    facsicular_state       : dict
    """

    fascicular = json_load(dir_path + "00_Fascicle_config.json")
    facsicular_state = {"-1": fascicular}
    N_ax = fascicular["N"]
    for i in range(N_ax):
        file = "sim_axon_"+ str(i) + ".json"
        if "extra_stim" not in facsicular_state["-1"]:
            facsicular_state["-1"]["extra_stim"] = extra_stim_properties(
                dir_path + file
            )

        axon = axon_state(dir_path + file)
        facsicular_state[axon["ID"]] = axon

        if delete_files:
            rm_file(dir_path + file, verbose)
    if save:
        save_axon_results_as_json(facsicular_state, saving_file)

    return facsicular_state


#############################
## VISUALIZATION FUNCTIONS ##
#############################


def plot_fasc_state(
    facsicular_state,
    fig,
    axes,
    contour_color="k",
    myel_color="r",
    unmyel_color="b",
    num=False,
):
    """
    plot the fascicle in the Y-Z plane (transverse section)

    Parameters
    ----------
    fig     : matplotlib.figure
        figure to display the fascicle
    axes    : matplotlib.axes
        axes of the figure to display the fascicle
    contour_color   : str
        matplotlib color string applied to the contour. Black by default
    myel_color      : str
        matplotlib color string applied to the myelinated axons. Red by default
    unmyel_color    : str
        matplotlib color string applied to the myelinated axons. Blue by default
    num             : bool
        if True, the index of each axon is displayed on top of the circle
    """

    fasc = facsicular_state["-1"]
    colors = []
    alpha = []
    N = 0
    while N in facsicular_state:
        if facsicular_state[N]["block_state"]:
            if facsicular_state[N]["onset_state"]:
                colors += ["green"]
                alpha += [facsicular_state[N]["onset number"]]
            else:
                colors += ["bleu"]
                alpha += [0]
        elif facsicular_state[N]["block_state"] is None:
            colors += ["red"]
            alpha += [facsicular_state[N]["onset number"]]
        else:
            if facsicular_state[N]["onset_state"]:
                colors += ["orange"]
                alpha += [facsicular_state[N]["onset number"]]
            else:
                colors += ["lightgray"]
                alpha += [0]
        N += 1

    alpha = [1 - (i / (1 + max(alpha))) for i in alpha]

    ## plot contour
    axes.plot(fasc["y_vertices"], fasc["z_vertices"], linewidth=2, color=contour_color)
    ## plot axons
    circles = []
    for k in range(N):
        if fasc["axons_type"][k] == 1:  # myelinated
            circles.append(
                plt.Circle(
                    (fasc["axons_y"][k], fasc["axons_z"][k]),
                    fasc["axons_diameter"][k] / 2,
                    color=colors[k],
                    fill=True,
                    alpha=alpha[k],
                )
            )
        else:
            circles.append(
                plt.Circle(
                    (fasc["axons_y"][k], fasc["axons_z"][k]),
                    fasc["axons_diameter"][k] / 2,
                    color=colors[k],
                    fill=True,
                    alpha=alpha[k],
                )
            )
    for circle in circles:
        axes.add_patch(circle)
    if num:
        for k in range(N):
            axes.text(fasc["axons_y"][k], fasc["axons_z"][k], str(k))
    ## plot electrode(s) if existings
    if "extra_stim" in fasc:
        extra_stim = load_any(fasc["extra_stim"])
        for electrode in extra_stim.electrodes:
            if electrode.type == "LIFE":
                circles.append(
                    plt.Circle(
                        (electrode.y, electrode.z),
                        electrode.D / 2,
                        color="gold",
                        fill=True,
                    )
                )
            elif not is_FEM_electrode(electrode):
                axes.scatter(electrode.y, electrode.z, color="gold")


"""
NRV-Cellular Level discretization methods
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS


todo:
 - Blocking vs dt
 - Blocking vs nseg
 - Onset_response vs dt (ou pas)
 - Onset_response vs dt (ou pas)
 - Velocity vs nseg 
 - Spike th vs nseg unmyelinated

"""
import numpy as np
from scipy.optimize import minimize

from ...backend.log_interface import rise_error, rise_warning

calibrated_unmyelinated_models = ["Rattay_Aberham", "Sundt", "Tigerholm"]
calibrated_myelinated_models = ["MRG", "Gaines_motor", "Gaines_sensory"]
# calibrated_simulation = ['Velocity','Spike_threshold','Blocking_threshold','Onset_response']
calibrated_simulation = ["Velocity", "Spike_threshold"]


def _diff_nseg_spike_threshold_myelinated(nseg, tol):
    """
    Computes the norm to minimize for a specified nseg - Spike Threshold Simulation - Myelinated models

    Parameters
    ----------
    nseg : int
        number of segment

    tol:
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """
    nseg_log = np.log(nseg)
    tol_log = np.log(tol)
    poly_coeff = [0.08914346, -1.2569447, 3.12523366]
    poly = np.poly1d(poly_coeff)
    val = np.exp(poly(nseg_log))
    return (val - tol) ** 2


def _diff_dt_ratio_spike_threshold_myelinated(dt_ratio, tol):
    """
    Computes the norm to minimize for a specified dt - Spike Threshold Simulation - Myelinated models

    Parameters
    ----------
    dt_ratio : float
        dt ratio

    tol:
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """
    dt_log = np.log(dt_ratio)
    tol_log = np.log(tol)
    poly_coeff = [-0.05832062, 0.95287425, -5.45710455, 12.26637458, -7.04848512]
    poly = np.poly1d(poly_coeff)
    val = np.exp(poly(dt_log))
    return (val - tol) ** 2


def _diff_dt_ratio_spike_threshold_unmyelinated(dt_ratio, tol):
    """
    Computes the norm to minimize for a specified dt - Spike Threshold Simulation - Unmyelinated models

    Parameters
    ----------
    dt_ratio : float
        dt ratio

    tol:
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """
    dt_log = np.log(dt_ratio)
    tol_log = np.log(tol)
    poly_coeff = [0.01751393, -0.25108173, 1.23328312, -3.34147125, 6.57882033]
    poly = np.poly1d(poly_coeff)
    val = np.exp(poly(dt_log))
    return (val - tol) ** 2


def _diff_dt_velocity_myelinated(dt, tol):
    """
    Computes the norm to minimize for a specified dt - Velocity estimation - Myelinated models

    Parameters
    ----------
    dt : float
        time discretization in ms

    tol:
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """
    poly_coeff = [3.65238002e05, -4.13999358e04, 2.16625888e03, -2.75739556e-01]
    poly = np.poly1d(poly_coeff)
    return (poly(dt) - tol) ** 2


def _diff_dt_velocity_sundt_tigerholm(dt, tol):
    """
    Computes the norm to minimize for a specified dt - Velocity estimation - Sundt-Tigerholm unmyelinated models

    Parameters
    ----------
    dt : float
        time discretization in ms

    tol:
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """

    poly_coeff = [1.70847637e05, -1.40276023e04, 5.26842025e02, 1.27772606e-01]
    poly = np.poly1d(poly_coeff)
    return (poly(dt) - tol) ** 2


def _diff_dt_velocity_rattay_aberham(dt, tol):
    """
    Computes the norm to minimize for a specified dt - Velocity estimation - Rattay-Aberham unmyelinated models

    Parameters
    ----------
    dt : float
        time discretization in ms

    tol:
        desired tolerance

    Returns
    -------
    float
        norm to minimize
    """

    poly_coeff = [2.79799924e05, -1.18439668e04, 7.14708491e02, -1.43308070e-01]
    poly = np.poly1d(poly_coeff)
    return (poly(dt) - tol) ** 2


def _get_dt_velocity_myelinated(tol, model):
    """
    returns the dt value for requested tolerance - myelinated models - velocity estimation

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt
        time discretization in ms
    """
    if tol > 50:
        rise_warning(
            "Requested tolerance is out of calibration range. Returned value might be inacurate."
        )
    res = minimize(
        _diff_dt_velocity_myelinated, 1.0, args=(tol), method="Nelder-Mead", tol=1e-6
    )
    return res.x[0]


def _get_dt_velocity_unmyelinated(tol, model):
    """
    returns the dt value for requested tolerance - unmyelinated models - velocity estimation

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt
        time discretization in ms
    """
    if model == "Rattay_Aberham":
        if tol > 40:
            rise_warning(
                "Requested tolerance is out of calibration range. Returned value might be inacurate."
            )
        res = minimize(
            _diff_dt_velocity_rattay_aberham,
            1.0,
            args=(tol),
            method="Nelder-Mead",
            tol=1e-6,
        )
    if (model == "Sundt") or (model == "Tigerholm"):
        if tol > 10:
            rise_warning(
                "Requested tolerance is out of calibration range. Returned value might be inacurate."
            )
        res = minimize(
            _diff_dt_velocity_sundt_tigerholm,
            1.0,
            args=(tol),
            method="Nelder-Mead",
            tol=1e-6,
        )
    return res.x[0]


def _get_dt_spike_threshold_myelinated(tol, model):
    """
    returns the dt value for requested tolerance - myelinated models - Spike threshold

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt_ratio
        time discretization ratio
    """
    if tol > 10:
        rise_warning(
            "Requested tolerance is out of calibration range. Returned value might be inacurate."
        )
    res = minimize(
        _diff_dt_ratio_spike_threshold_myelinated,
        10.0,
        args=(tol),
        method="Nelder-Mead",
        tol=1e-6,
    )
    return res.x[0]


def _get_dt_spike_threshold_unmyelinated(tol, model):
    """
    returns the dt value for requested tolerance - unmyelinated models - Spike threshold

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt_ratio
        time discretization ratio
    """
    if tol > 40:
        rise_warning(
            "Requested tolerance is out of calibration range. Returned value might be inacurate."
        )
    res = minimize(
        _diff_dt_ratio_spike_threshold_unmyelinated,
        1.0,
        args=(tol),
        method="Nelder-Mead",
        tol=1e-6,
    )
    return res.x[0]


def _get_nseg_spike_threshold_myelinated(tol, model):
    """
    returns the number of segment for the requested tolerance - myelinated models - Spike threshold

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    nseg: int
        number of segment
    """
    if tol > 25:
        rise_warning(
            "Requested tolerance is out of calibration range. Returned value might be inacurate."
        )
    res = minimize(
        _diff_nseg_spike_threshold_myelinated,
        1.0,
        args=(tol),
        method="Nelder-Mead",
        tol=1e-6,
    )
    return res.x[0]


def _get_dt_velocity(tol, model):
    """
    returns the dt value for requested tolerance - Velocity estimation

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt
        time discretization in ms
    """

    if model in calibrated_myelinated_models:
        return _get_dt_velocity_myelinated(tol, model)
    if model in calibrated_unmyelinated_models:
        return _get_dt_velocity_unmyelinated(tol, model)
    if model not in (calibrated_myelinated_models) and model not in (
        calibrated_unmyelinated_models
    ):
        rise_error("Selected model is not calibrated yet.")
        return 0


def _get_dt_spike_threshold(tol, model):
    """
    returns the dt value for requested tolerance - Spike threshold

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model
    Returns
    -------
    dt
        time discretization ratio
    """

    if model in calibrated_myelinated_models:
        return _get_dt_spike_threshold_myelinated(tol, model)
    if model in calibrated_unmyelinated_models:
        return _get_dt_spike_threshold_unmyelinated(tol, model)
    if model not in (calibrated_myelinated_models) and model not in (
        calibrated_unmyelinated_models
    ):
        rise_error("Selected model is not calibrated yet.")


def _get_nseg_spike_threshold(tol, model, L=None, d=None):
    """
    returns the number of segment for requested tolerance - Spike threshold

    Parameters
    ----------
    tol : float
        tolerance

    model: str
        simulated model

    d: float
        axon diameter in um (only for unmyelinated models)

    L: float
        Axon length in um (only for unmyelinated models)
    Returns
    -------
    nseg
        number of segment
    """

    if model in calibrated_myelinated_models:
        return _get_nseg_spike_threshold_myelinated(tol, model)
    if model in calibrated_unmyelinated_models:
        rise_error("Not implemented yet.")
    if model not in (calibrated_myelinated_models) and model not in (
        calibrated_unmyelinated_models
    ):
        rise_error("Selected model is not calibrated yet.")


def Get_nseg(tol, model, simulation_category, d=None, L=None):
    """
    Returns the number of segment for the requested tolerance

    Parameters
    ----------
    tol : float
        desired tolerance

    model: str
        simulated

    simulation_category: str
        Simulation to perform

    d: float
        axon diameter in um (only for unmyelinated models)

    L: float
        axon length in um (only for unmyelinated models)

    Returns
    -------
    nseg number of segments
    """

    if simulation_category not in (calibrated_simulation):
        rise_error("Selected simulation is not calibrated yet.")
    else:
        if simulation_category == "Velocity":
            rise_error("Not implemented yet.")

        if simulation_category == "Spike_threshold":
            nseg = _get_nseg_spike_threshold(tol, model)
            return np.int32(np.round(nseg))


def Get_dt(tol, model, simulation_category, stim_pw=None):
    """
    Estimates the dt value for requested tolerance

    Parameters
    ----------
    tol : float
        desired tolerance

    model: str
        simulated

    simulation_category: str
        Simulation to perform

    stim_pw: float
        Stimulation characteristic pulse width in ms (not required for velocity simulation)

    Returns
    -------
    dt
        time discretization value
    """

    if simulation_category not in (calibrated_simulation):
        rise_error("Selected simulation is not calibrated yet.")
    else:
        if simulation_category == "Velocity":
            return _get_dt_velocity(tol, model)

        if simulation_category == "Spike_threshold":
            if stim_pw > 250e-3:
                rise_warning(
                    "Characteristic pulse width is out of calibration range. Returned value might be inacurate."
                )
            dt_ratio = stim_pw / _get_dt_spike_threshold(tol, model)
            return dt_ratio



"""
NRV-Stimulus
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler

import numpy as np

from ..backend.log_interface import pass_info, rise_warning
from ..backend.NRV_Class import NRV_class

# enable faulthandler to ease "segmentation faults" debug
faulthandler.enable()


###############
## Functions ##
###############
def is_stim(stim):
    """
    check if an object is a stimulus, return True if yes, else False

    Parameters
    ----------
    stim : object
        object to test

    Returns
    -------
    bool
        True it the type is a stimulus object
    """
    return isinstance(stim, stimulus)


def set_common_time_series(stim1, stim2):
    """
    set two signals at the same time samples, missing samples are reconstructed by holding a value

    Parameters
    ----------
    stim1   : stimulus object
        First stimulus to synchronize
    stim2   : stimulus object
        Second object to synchronize
    """
    stim1.insert_samples(stim2.t)
    stim2.insert_samples(stim1.t)


def get_equal_timing_copies(stim1, stim2):
    """
    create two stimuli with equal timing as copies of the input ones

    Parameters
    ----------
    stim1   : stimulus object
        First stimulus to synchronize
    stim2   : stimulus object
        Second object to synchronize

    Returns
    --------
    stima : stimulus object
        copie of stim1 with timings from stim1 and stim2
    stimb : stimulus object
        copie of stim2 with timings from stim1 and stim2
    """
    stim_a = stimulus()
    stim_a.s = stim1.s
    stim_a.t = stim1.t
    stim_b = stimulus()
    stim_b.s = stim2.s
    stim_b.t = stim2.t
    stim_a.insert_samples(stim_b.t)
    stim_b.insert_samples(stim_a.t)
    return stim_a, stim_b


def datfile_2_stim(fname, dt=0.005):
    """Creates a stimulus from sampled data specified in a .dat file

    Parameters
    ----------
    fname : str
        name of the file containing the sampled data
    dt    : float, optional
        value of the sampling period, in ms. By default set to 0.005ms
    """
    stim_1 = stimulus()
    # open the dat file and feed the stimulus
    data = np.genfromtxt(fname)
    stim_1.t = np.arange(len(data)) * dt
    stim_1.s = data
    return stim_1


################################
## Stimulus object definition ##
################################
class stimulus(NRV_class):
    """
    Stimulus class for NRV2,
    signals are defined as asynchronous signals, with s the values and t as occurence timings.
    """

    def __init__(self, s_init=0):
        """
        Instantiation of a stimulus object.

        Parameters
        ----------
        s_init  : float
            initial value of the signal, by default 0
        """
        super().__init__()
        self.type = "stimulus"
        self.s = np.array([s_init])
        self.t = np.array([0])

    ## Save and Load mehtods
    def save_stimulus(self, save=False, fname="stimulus.json"):
        rise_warning("save_stimulus is a deprecated method use save")
        self.save(save=save, fname=fname)

    def load_stimulus(self, data="stimulus.json"):
        rise_warning("load_stimulus is a deprecated method use load")
        self.load(data=data)

    ############################
    ## basic handling methods ##
    ############################
    def append(self, s, t):
        """
        Append a sample to the signal, internal use only.

        Parameters
        ----------
        s   : float
            value of the signal to append, numpy array like
        t   : float
            time of signal change
        """
        self.s = np.append(self.s, np.asarray(s))
        self.t = np.append(self.t, np.asarray(t))

    def concatenate(self, s, t, t_shift=0):
        """
        Concatenate samples to a signal, internal use only.

        Parameters
        ----------
        s       : float
            values of signal, numpy array like
        t       : float
            time serie of the values, numpy array like
        t_shift : float
            time to shift the t values at the end of original signal, by default equal to 0.
            Here to prevent possible duplicates if the t serie starts with 0
        """
        self.s = np.append(self.s, np.asarray(s))
        self.t = np.append(self.t, np.asarray(t) + self.t[-1] + t_shift)

    def len(self):
        """
        Returns the length of the signal

        Returns
        -------
        len : int
            length of the signal (s and t attribute should have the same length
            if the signal is not corrupted)
        """
        return len(self.s)

    def sort(self):
        """
        Sort the signal with increasing timings
        """
        ordered_ind = np.argsort(self.t)
        self.s = self.s[ordered_ind]
        self.t = self.t[ordered_ind]

    def insert_samples(self, t):
        """
        Insert samples inside a signal and adapt values consequently

        Parameters
        ----------
        t   : array, np.array or list
            time serie of samples to insert, numpy array-like, dosent have to be sorted
        """
        # add the new time values at their place and remove duplicates
        new_t = np.unique(np.sort(np.append(self.t, np.asarray(t))))
        new_s = [self.s[0]]
        j = 1
        for k in range(1, len(new_t)):
            if j < self.len():
                if new_t[k] < self.t[j]:  # added sample, hold the last value
                    new_s.append(new_s[-1])
                else:  # orinal sample, retrieve value in original signal and increment counter
                    new_s.append(self.s[j])
                    j += 1
            else:
                new_s.append(self.s[-1])
        self.s = np.asarray(new_s)
        self.t = new_t

    def snap_time(self, dt_min):
        i_mask = [True for _ in range(len(self.t))]
        for i in range(len(self.t) - 1):
            j = 0
            if self.t[i + 1] - self.t[i] < dt_min:
                while i > j and not i_mask[i - j]:
                    j += 1
                if self.t[i + 1] - self.t[i - j] < dt_min:
                    i_mask[i + 1] = False
        self.s = self.s[i_mask]
        self.t = self.t[i_mask]

    #####################
    ## special methods ##
    #####################
    def __len__(self):
        return self.len()

    def __abs__(self):
        stim = stimulus()
        stim.s = np.absolute(self.s)
        stim.t = self.t
        return stim

    def __neg__(self):
        stim = stimulus()
        stim.s = -self.s
        stim.t = self.t
        return stim

    def __add__(self, b):
        C = stimulus()
        if is_stim(b):
            A, B = get_equal_timing_copies(self, b)
            C.s = A.s + B.s
            C.t = A.t
        else:
            C.s = self.s + float(b)
            C.t = self.t
        return C

    def __sub__(self, b):
        C = stimulus()
        if is_stim(b):
            A, B = get_equal_timing_copies(self, b)
            C.s = A.s - B.s
            C.t = A.t
        else:
            C.s = self.s - float(b)
            C.t = self.t
        return C

    def __mul__(self, b):
        C = stimulus()
        if is_stim(b):
            A, B = get_equal_timing_copies(self, b)
            C.s = A.s * B.s
            C.t = A.t
        else:
            C.s = self.s * float(b)
            C.t = self.t
        return C

    def __radd__(self, b):
        C = stimulus()
        if is_stim(b):
            A, B = get_equal_timing_copies(self, b)
            C.s = B.s + A.s
            C.t = A.t
        else:
            C.s = float(b) + self.s
            C.t = self.t
        return C

    def __rsub__(self, b):
        C = stimulus()
        if is_stim(b):
            A, B = get_equal_timing_copies(self, b)
            C.s = B.s - A.s
            C.t = A.t
        else:
            C.s = float(b) - self.s
            C.t = self.t
        return C

    def __rmul__(self, b):
        C = stimulus()
        if is_stim(b):
            A, B = get_equal_timing_copies(self, b)
            C.s = B.s * A.s
            C.t = A.t
        else:
            C.s = float(b) * self.s
            C.t = self.t
        return C

    def __pow__(self, p):
        C = stimulus()
        C.s = self.s ** float(p)
        return C

    def __eq__(self, b):
        if is_stim(b):
            A, B = get_equal_timing_copies(self, b)
            flag = np.array_equal(A.s, B.s) and np.array_equal(A.t, B.t)
        else:
            flag = True
            for value in self.s:
                if value != b:
                    flag = False
        return flag

    def __ne__(self, b):  # self != b
        return not self == b


    def integrate(self):
        return np.trapz(self.s, x=self.t)

    #######################
    ## signal generators ##
    #######################
    def constant(self, value, start=0):
        """
        Ceat a constant signal

        Parameters
        ----------
        value : float
            Value of the constant signal
        start :float
            starting time of the constant signal, by default set to 0
        """
        self.append(value, start)

    def pulse(self, start, value, duration=0):
        """
        Create pulse shape signal samples

        Parameters
        ----------
        value       : float
            value of the pulse
        start       : float
            starting time of the pulse
        duration    : float
            duration of the pulse, optional and by default equal to zero, if non zero value,
            a point is added at the end of the pulse with the previous value
        """
        s_last = self.s[-1]
        self.append(value, start)
        if duration != 0:
            self.append(s_last, start + duration)

    def biphasic_pulse(
        self, start, s_cathod, t_stim, s_anod, t_inter, anod_first=False
    ):
        """
        Create a biphasic pulse waveform

        Parameters
        ----------
        start       : float
            starting time of the waveform, in ms
        s_cathod    : float
            cathodic (negative stimulation value) current, in uA
            WARNING: always positive, the user give here the absolute value
        t_stim      : float
            stimulation time, in ms
        s_anod      : float
            anodic (positive stimulation value) current and, in uA
        t_inter     : float
            inter pulse timing, in ms
        anod_first  : bool
            if true, stimulation is anodic and begins with the anodic value
            and is balanced with cathodic value, else stimuation is cathodic
            and begins with the cathodic value and is balances with anodic value,
            by default set to False (cathodic first as most stimulation protocols)
        """
        if not anod_first:
            s_1 = -s_cathod
            s_2 = s_anod
        else:
            s_2 = -s_cathod
            s_1 = s_anod
        if s_2 != 0:
            t_balance = abs(s_1 / s_2) * t_stim
        else:
            t_balance = 0
        self.concatenate(s_1, start)
        if t_inter == 0:
            self.concatenate(s_2, t_stim)
            if t_balance != 0:
                self.concatenate(0, t_balance)
        else:
            self.concatenate(0, t_stim)
            self.concatenate(s_2, t_inter)
            if t_balance != 0:
                self.concatenate(0, t_balance)

    def sinus(self, start, duration, amplitude, freq, offset=0, phase=0, dt=0):
        """
        Create a sinusoidal waveform

        Parameters
        ----------
        start       : float
            starting time of the waveform, in ms
        duration    : float
            duration of the waveform, in ms
        amplitude   : float
            amplitude of the waveform, in uA
        freq        : float
            frequency of the waveform, in kHz
        offset      : float
            offset current of the waveform, in uA, by default set to 0
        phase       : float
            initial phase of the waveform, in rad, by default set to 0
        dt          : float
            sampling time period to generate the sinusoidal shape. If equal to 0,
            dt is automatically set to match 100 samples per sinusoid period by default set to 0
        """
        # check the pseudo sampling period
        if dt == 0:
            dt = 1 / (freq * 100)
        elif freq > (1.0 / (2 * dt)):
            rise_warning(
                "dt too low in stimulus creation, Shannon criterion not respected"
            )
        Nb_points = int(duration / dt)
        # create the signal
        if start == 0:
            self.s[0] = amplitude * np.sin(phase) + offset
            t = np.linspace(dt, start + duration, num=Nb_points - 1)
            s = amplitude * np.sin(2 * np.pi * freq * t + phase) + offset
            self.concatenate(s, t, t_shift=0)
        else:
            t = np.linspace(start, start + duration, num=Nb_points)
            s = amplitude * np.sin(2 * np.pi * freq * t + phase) + offset
            self.concatenate(s, t, t_shift=0)

    def square(self, start, duration, freq, amplitude, offset, dt):
        """
        Create a repetitive (periodic) square waveform

        Parameters
        ----------
        start       : float
            starting time of the waveform, in ms
        duration    : float
            duration of the waveform, in ms
        amplitude   : float
            amplitude of the waveform, in uA
        freq        : float
            frequency of the waveform, in kHz
        offset      : float
            offset current of the waveform, in uA, by default set to 0
        dt          : float
            sampling time period to generate the sinusoidal shape. If equal to 0,
            dt is automatically set to match 100 samples per sinusoid period by default set to 0
        """
        Nb_points = int(duration / dt)
        if start == 0:
            Nb_points -= 1
        T = 1 / freq
        t = np.linspace(dt, start + duration, num=Nb_points)
        s = np.ones(Nb_points)
        point_start = int(start / dt)
        for i in range(Nb_points):
            if i < point_start:
                s[i] = 0
            else:
                if t[i] % T < T / 2:
                    s[i] = -s[i]
                s[i] = amplitude * s[i] + offset
        self.concatenate(s, t, t_shift=0)

    def ramp(
        self, slope, start, duration, dt, bounds=(0, float("inf")), printslope=False
    ):
        """
        Create a ramp waveform with slop value

        Parameters
        ----------
        slope       : float
            slope of the waveform, in uA.ms-1
        start       : float
            starting time of the waveform, in ms
        duration    : float
            duration of the waveform, in ms
        dt          : float
            sampling time period to generate the sinusoidal shape
        bounds      : tuple
            boundary vaues of the ramp signal
        printslope  : bool, optional
            if True, the value of the slope is printed on the prompt
        """
        # create the signal
        if printslope:
            pass_info("slope = " + str(slope))
        Nb_points = int(duration / dt)
        t = np.array([k * dt for k in range(Nb_points)])
        if slope < 0:
            s = max(bounds) * np.ones(Nb_points)
        else:
            s = min(bounds) * np.ones(Nb_points)
        point_start = int(start / dt)
        for i in range(point_start, Nb_points):
            if slope >= 0:
                s[i] = min(bounds[1], bounds[0] + (i - point_start) * slope * dt)
            if slope < 0:
                s[i] = max(bounds[0], bounds[1] + (i - point_start) * slope * dt)
        self.concatenate(s, t, t_shift=0)

    def ramp_lim(self, tstart, tstop, ampstart, ampmax, duration, dt, printslope=False):
        """
        Create a ramp waveform with bounds values

        Parameters
        ----------
        ampstart    : float
            initiale amplitude of the waveform, in uA
        ampmax      : float
            final amplitude of the waveform, in uA
        tstart      : float
            starting time of the waveform, in ms
        tmax        : float
            starting time of the waveform, in ms
        duration    : float
            duration of the waveform, in ms
        dt          : float
            sampling time period to generate the sinusoidal shape
        printslope  : bool, optional
            if True, the value of the slope is printed on the prompt
        """
        slope = (ampstart - ampmax) / (tstart - tstop)
        bounds = (min(ampstart, ampmax), max(ampstart, ampmax))
        self.ramp(slope, tstart, duration, dt, bounds, printslope)






"""
NRV-FEM
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""

import configparser
import os
import time

import mph
import numpy as np

from ...backend.file_handler import rmv_ext
from ...backend.MCore import MCH
from ...backend.log_interface import rise_warning
from ...utils.units import V
from .FEM import *

# built in COMSOL models
dir_path = os.environ["NRVPATH"] + "/_misc"
material_library = os.listdir(dir_path + "/comsol_templates/")

###############
## Constants ##
###############
machine_config = configparser.ConfigParser()
config_fname = dir_path + "/NRV.ini"
machine_config.read(config_fname)
COMSOL_Ncores = machine_config.get("COMSOL", "COMSOL_CPU")
COMSOL_Status = machine_config.get("COMSOL", "COMSOL_STATUS") == "True"

fem_verbose = True


class COMSOL_model(FEM_model):
    """
    A class for COMSOL Finite Element Models, inherits from FEM_model.
    """

    def __init__(self, fname, Ncore=None, handle_server=False):
        """
        Creates a instance of a COMSOL Finite Element Model object.

        Parameters
        ----------
        fname   : str
            path to the COMSOL (.mph) model file
        Ncore   : int
            number of COMSOL cores for computation. If None is specified, this number is taken from the NRV2.ini configuration file. Byu default set to None
        handle_server   : bool
            if True, the instantiation creates the server, else a external server is used. Usefull for multiple cells sharing the same model
        """
        if COMSOL_Status:
            t0 = time.time()
            super().__init__(Ncore=Ncore)
            self.type = "COMSOL"

            self.model_path = fname
            f_in_librairy = rmv_ext(str(fname)) + ".mph"
            if f_in_librairy in material_library:
                self.fname = dir_path + "/comsol_templates/" + f_in_librairy
            else:
                self.fname = fname
            # self.model_path = fname
            if self.Ncore is None:
                self.Ncore = COMSOL_Ncores
            self.is_multi_proc = self.Ncore > 1 and not MCH.is_alone()
            self.handle_server = handle_server
            # start client and server
            pass_info("Starting COMSOL server/client, this may take few seconds")
            if self.handle_server:
                self.server = mph.Server(cores=self.Ncore)
            else:
                self.server = None
            self.client = mph.start(cores=self.Ncore)
            self.client.caching(True)
            pass_info("... loading the COMSOL model")
            self.model = self.client.load(self.fname)
            # self.client.caching(True)
            # source
            self.fname = fname
            self.preparing_timer += time.time() - t0
        else:
            ## COMSOL TURNED OFF, no error, but only a warning and no computation (exit)
            rise_warning(
                "Bad implementation, a COMSOL simulation is implemented while NRV2 has COMSOL status turned OFF, unterminated computation, early exit without error",
                abort=True,
            )

    def save(self):
        """
        Save the changes to the model file. (Avoid for the overal weight of the package)
        """
        self.model.save()

    def clear(self):
        """
        Clear the mesh and result section of the model
        """
        self.model.clear()
        self.model.reset()

    def close(self):
        """
        Close the FEM simulation and the COMSOL link
        """
        # self.client.disconnect()
        del self.client
        if self.handle_server:
            self.server.stop()
            del self.server

    #############################
    ## Access model parameters ##
    #############################
    def get_parameters(self):
        """
        Get the  all the parameters in the model as a python dictionary.

        Returns
        -------
        parameters
            all parameters as dictionnary, names a keys, with corresponding values
        """
        return self.model.parameters()

    def get_parameter(self, p_name):
        """
        Get a specific parameter

        Returns
        -------
        str
            value of the parameter as in COMSOL (with unit)
        """
        return self.model.parameter(p_name)

    def set_parameter(self, p_name, p_value):
        """
        Set a parameter to a desired value

        Parameters
        ----------
        p_name  : str
            parameter name in the COMSOL model
        p_value : str
            parameter value as in COMSOL, with unit
        """
        self.model.parameter(p_name, p_value)

    ###################
    ## Use the model ##
    ###################
    def get_meshes(self):
        """
        Get the different meshes implemented in the model

        Returns
        -------
        list
            list of meshes implemented in the COMSOL model file
        """
        return self.model.meshes()

    def build_and_mesh(self):
        """
        Build the geometry and perform meshing process
        """
        t0 = time.time()
        pass_info("... Building model geometry")
        self.model.build()
        pass_info("... Meshing geometry")
        self.model.mesh()
        self.is_meshed = True
        self.meshing_timer += time.time() - t0

    def solve(self):
        """
        Solve the model
        """
        if not self.is_meshed:
            self.build_and_mesh()
        pass_info("... Solving model")
        t0 = time.time()
        self.model.solve()
        self.is_computed = True
        self.solving_timer += time.time() - t0

    ######################
    ## results handling ##
    ######################
    def get_potentials(self, x, y, z):
        """
        Get the potential on a line to get extracellular potential for axons stimulation.

        Parameters
        ----------
        x   : np.array
            array of x coordinates in the model
        y   : float
            y-coordinate of the axon
        z   : float
            z-coordinate of the axon

        Returns:
        --------
        array
            All potential for all paramtric sweeps (all electrodes in NRV2 models)\
            (line: electrode selection, column: potential)
        """
        t0 = time.time()
        COMSOL_expressions = [
            "at3(" + str(x[k]) + "[um], " + str(y) + "[um], " + str(z) + "[um], V)"
            for k in range(len(x))
        ]
        Voltage = self.model.evaluate(COMSOL_expressions) * V
        self.access_res_timer += time.time() - t0
        return np.asarray(Voltage)

    def export(self, path=""):
        """
        Export the figures of the COMSOL results and posprocess (in PNG format)

        Parameters
        ----------
        path    : str
            path address where to save graphics
        """
        exports = self.model.exports()
        for export in exports:
            self.model.export(export, path + export + ".png")



import math
import os

import gmsh
import numpy as np

from ....backend.file_handler import rmv_ext
from ....backend.log_interface import pass_info, rise_error, rise_warning
from ....backend.NRV_Class import NRV_class
from ....backend.parameters import parameters

dir_path = os.environ["NRVPATH"] + "/_misc"


###############
## Constants ##
###############

pi = np.pi


def is_MshCreator(object):
    """
    Check if an object is a MshCreator, return True if yes, else False

    Parameters
    ----------
    result : object
        object to test

    Returns
    -------
    bool
        True it the type is a MshCreator object
    """
    return isinstance(object, MshCreator)


def clear_gmsh():
    if gmsh.isInitialized():
        gmsh.finalize()


class MshCreator(NRV_class):
    """
    Class handeling the creation of a gmsh mesh (https://gmsh.info/doc/texinfo/gmsh.html)
    Contains methodes dealing with the mesh geometries, physical domains and feilds
    """

    def __init__(self, D=3, ver_level=None):
        """
        initialisation of the MshCreator

        Parameters
        ----------
        D           : int(1,2,3)
            mesh dimension
        ver_level   : int(0,1,2,3,4,5,99)
            verbosity level of gmsh (see MshCreator.set_verbosity), by default 2
        """
        super().__init__()
        self.type = "GMSH"
        self.D = D
        self.entities = {}
        self.volumes = []
        self.volumes_com = []
        self.volumes_bd = []

        self.faces = []
        self.faces_com = []
        self.faces_bd = []

        self.N_domains = 0
        self.domains = []
        self.file = ""

        self.Nfeild = 0
        self.res = 1

        clear_gmsh()
        gmsh.initialize()

        self.verbosity_level = ver_level
        self.set_verbosity(ver_level)
        self.model = gmsh.model
        self.model.add("DFG 3D")

        # Mesh properties
        self.is_generated = False
        self.N_entities = 0
        self.N_nodes = 0
        self.N_elements = 0

        self.Ncore = parameters.GMSH_Ncores
        gmsh.option.setNumber("General.NumThreads", self.Ncore)
        if self.Ncore > 1:
            gmsh.option.set_number("Mesh.Algorithm3D", 10)
        else:
            gmsh.option.set_number("Mesh.Algorithm3D", 1)

    #####################
    ## special methods ##
    #####################

    def get_obj(self):
        """
        update and return list of mesh entities

        Returns
        -------
        self.entities       :dict
        """
        return self.entities

    def get_volumes(self, com=False, bd=False):
        """
        update and return list of mesh volumes (optional: with their center of mass)

        Returns
        -------
        self.faces      :list[tuple]
        """
        self.model.occ.synchronize()
        self.volumes = self.model.getEntities(dim=3)
        self.volumes_com = []
        self.volumes_bd = []
        for i in range(len(self.volumes)):
            volume = self.volumes[i]
            self.volumes_com += [
                np.round(self.model.occ.getCenterOfMass(volume[0], volume[1]), 4)
            ]
            self.volumes_bd += [
                np.round(self.model.occ.getBoundingBox(volume[0], volume[1]), 4)
            ]

        if com:
            if bd:
                return self.volumes, self.volumes_com, self.volumes_bd
            else:
                return self.volumes, self.volumes_com
        else:
            if bd:
                return self.volumes, self.volumes_bd

        return self.volumes

    def get_info(self, verbose=False):
        entities = self.model.getEntities()
        nodeTags = self.model.mesh.getNodes()[0]
        elemTags = self.model.mesh.getElements()[1]

        self.N_entities = len(entities)
        self.N_nodes = len(nodeTags)
        self.N_elements = sum(len(i) for i in elemTags)
        if verbose:
            pass_info("Mesh properties:")
            pass_info("Number of processes : " + str(self.Ncore))
            pass_info("Number of entities : " + str(self.N_entities))
            pass_info("Number of nodes : " + str(self.N_nodes))
            pass_info("Number of elements : " + str(self.N_elements))
        return self.Ncore, self.N_entities, self.N_nodes, self.N_elements

    def get_mesh_info(self, verbose=False):
        rise_warning("DEPRECATED method use get_info instead")
        self.get_info(verbose=verbose)

    def get_faces(self, com=False, bd=False):
        """
        update and return list of mesh face (optional: with their center of mass)

        Returns
        -------
         self.faces     :list[tuple]
        """
        self.model.occ.synchronize()
        self.faces = self.model.getEntities(dim=2)
        self.faces_com = []
        self.faces_bd = []
        for i in range(len(self.faces)):
            face = self.faces[i]
            self.faces_com += [
                np.round(self.model.occ.getCenterOfMass(face[0], face[1]), 4)
            ]
            self.faces_bd += [
                np.round(self.model.occ.getBoundingBox(face[0], face[1]), 4)
            ]

        if com:
            if bd:
                return self.faces, self.faces_com, self.faces_bd
            else:
                return self.faces, self.faces_com
        else:
            if bd:
                return self.faces, self.faces_com
        return self.faces

    def get_res(self):
        """
        return the global resolution saved (usefull when no field are set)

        Returns
        -------
        res     :float
            global resolution saved in the object
        """
        return self.res

    def set_res(self, new_res):
        """
        set the global resolution saved (usefull when no field are set)

        Parameters
        ----------
        new_res     :float
            global resolution to set the object
        """
        self.res = new_res

    def set_verbosity(self, i=None):
        """
        from gmsh: Level of information printed on the terminal and the message console\
            
        Parameters
        ----------
        i     : int (1, 2, 3, 4, 5 or 99)
            0: silent except for fatal errors
            1: +errors
            2: +warnings
            3: +direct
            4: +information
            5: +status
            99: +debug
        """
        if i is None:
            i = parameters.VERBOSITY_LEVEL
        self.verbosity_level = i
        gmsh.option.setNumber("General.Verbosity", self.verbosity_level)

    def set_chara_blen(self, i=0):
        """
        from gmsh: Extend characteristic lengths from the boundaries inside the surface/volume

        Parameters
        ----------
        i   : int, float, bool
            Parameter value, by default 0
        """
        i = float(i)
        gmsh.option.set_number("Mesh.CharacteristicLengthExtendFromBoundary", i)

    ##############################################################################################
    #######################################   geometry methods  ##################################
    ##############################################################################################

    def add_point(self, x=0, y=0, z=0):
        """
        add a point to the mesh

        Parameters
        ----------
        x        : float
            x position of the first face center
        y        : float
            y position of the first face center
        z        : float
            z position of the first face center

        Returns
        -------
        point    : int
            id of the added object
        """
        point = self.model.occ.addPoint(x, y, z)
        self.model.occ.synchronize()
        return point

    def add_line(self, X0, X1):
        """
        add a line to the mesh

        Parameters
        ----------
        X0       : int or tupple(3)
            x position of the first face center
        X1       : int or tupple(3)
            y position of the first face center

        Returns
        -------
        line    : int
            id of the added object
        """
        if isinstance(X0, int):
            ix0 = X0
        elif np.iterable(X0):
            if len(X0) != self.D:
                rise_error("not enough dimension given in add_line")
            else:
                ix0 = self.add_point(X0[0], X0[1], X0[2])

        if isinstance(X1, int):
            ix0 = X0
        elif np.iterable(X1):
            if len(X1) != self.D:
                rise_error("not enough dimension given in add_line")
            else:
                ix1 = self.add_point(X1[0], X1[1], X1[2])
        line = self.model.occ.addLine(ix0, ix1)
        self.model.occ.synchronize()
        return line

    def add_box(self, x=0, y=0, z=0, ax=5, ay=1, az=1):
        """
        add a box to the mesh

        Parameters
        ----------
        x       : float
            x position of the first face center
        y       : float
            y position of the first face center
        z       : float
            z position of the first face center
        ax      : float
            Box length along x
        ay      : float
            Box length along y
        az      : float
            Box length along z

        Returns
        -------
        box    : int
            id of the added object
        """
        if self.D == 3:
            parameters = {"x": x, "y": y, "z": z, "ax": ax, "ay": ay, "az": az}
            box = self.model.occ.addBox(x, y, z, ax, ay, az)
            self.model.occ.synchronize()
            bounds = self.model.getEntities(dim=2)[-6:]
            self.entities[box] = {
                "type": "box",
                "parameters": parameters,
                "bounds": bounds,
                "dim": 3,
            }
            return box
        else:
            rise_warning("Not added : add_cylinder requiere 3D mesh")
            return None

    def add_cylinder(self, x=0, y=0, z=0, L=5, R=1):
        """
        add a x-oriented cylinder to mesh entities

        Parameters
        ----------
        x       : float
            x position of the first face center
        y       : float
            y position of the first face center
        z       : float
            z position of the first face center
        L       : float
            Cylinder length along x
        R       : float
            Cylinder radius

        Returns
        -------
        cyl    : int
            id of the added object
        """
        if self.D == 3:
            parameters = {"x": x, "y": y, "z": z, "L": L, "R": R}
            cyl = self.model.occ.addCylinder(x, y, z, L, 0, 0, R)
            self.model.occ.synchronize()
            bounds = self.model.getEntities(dim=2)[-3:]
            self.entities[cyl] = {
                "type": "cylinder",
                "parameters": parameters,
                "bounds": bounds,
                "dim": 3,
            }
            return cyl
        else:
            rise_warning("Not added : add_cylinder requiere 3D mesh")
            return None

    def rotate(self, volume, angle, x=0, y=0, z=0, ax=0, ay=0, az=0, rad=True):
        """
        rotate volume

        Parameters
        ----------
        volume      : int
            gmsh id of the volume
        angle:      : float
            angle of the rotation
        x           : float
            x-position of the center of the rotation, by default 0
        y           : float
            y-position of the center of the rotation, by default 0
        z           : float
            z-position of the center of the rotation, by default 0
        ax          : float
            x-coefficient of the rotation axis direction vector, by default 0
        ay          : float
            y-coefficient of the rotation axis direction vector, by default 0
        az          : float
            z-coefficient of the rotation axis direction vector, by default 0
        rad         : bool
            if true angle considered in rad, else in degree, by default 0
        """
        if not rad:
            angle = math.radians(angle)
        self.model.occ.rotate([(3, volume)], x, y, z, ax, ay, az, angle)

    def fragment(self, IDs=None, dim=3, verbose=True):
        """
        Fragmentation of the mesh important to link entities to each other

        Parameters
        ----------
        IDs     : list or None
            list of IDs of the element to fragments, if None all dim dimension elements are
            fragmented, by default None
        dim     : int
            dimension of the elements considerated, by default 3
        verbose : bool
            print or not the verbose on the temrminal, by default False
        """
        if IDs is None:
            if dim == 2:
                list_obj = self.get_faces(com=False)
            elif dim == 3:
                list_obj = self.get_volumes(com=False)
            else:
                list_obj = [(dim, k) for k in self.entities]
        elif not np.iterable(IDs) or len(IDs) < 2:
            rise_warning("Need at least 2 entities to fragment")
            return -1
        else:
            list_obj = [(dim, k) for k in IDs]

        frag = self.model.occ.fragment([list_obj[0]], [k for k in list_obj[1:]])
        new_entities = {}
        if verbose:
            pass_info(
                "Warning: New volume generated by fragmentation, bounds are no longer up to date"
            )
        for i in frag[0][:]:
            mask = [i in k for k in frag[1]]
            p_id = (np.array(list_obj)[mask][:, 1]).tolist()
            p_types = [self.entities[k]["type"] for k in p_id]
            dim = min([self.entities[k]["dim"] for k in p_id])
            com = np.round(self.model.occ.getCenterOfMass(i[0], i[1]), 3)
            parameters = {"p_id": p_id, "p_types": p_types, "com": com}
            new_entities[i[1]] = {
                "type": "fragment",
                "parameters": parameters,
                "dim": dim,
            }
        self.entities.update(new_entities)

    ##############################################################################################
    #######################################   domains methods  ###################################
    ##############################################################################################

    def add_domains(self, obj_IDs, phys_ID, dim=None, name=None):
        """
        add domains (ID + name) to a goupe of entities
        Caution: as to be used after all entities are placed

        Parameters
        ----------
        fname    : str
            path and name of saving file. If ends with ".msh" only save in ".msh" file
        """
        if not np.iterable(obj_IDs):
            obj_IDs = [obj_IDs]
        if dim is None:
            dim = max([self.entities[k]["dim"] for k in obj_IDs])
        self.model.addPhysicalGroup(dim, obj_IDs, phys_ID)

        if name is None:
            name = "domain " + str(self.N_domains)

        self.domains += (dim, phys_ID, name)

    ##############################################################################################
    #######################################   feilds methods  ####################################
    ##############################################################################################
    def refine_entities(self, ent_ID, res_in, dim, res_out=None, IncludeBoundary=True):
        """
        refine mesh resolution in a list of faces or volumes IDs

        Parameters
        ----------
        ent_ID    : list[int] or int
            ID or list of ID of the entities where the resolution should be changed
        res_in    : float
            resolution inside the entities
        dim    : int (2 or 3)
            dimention of the considerated entities
        res_out    : float
            resolution outside the entities if None take self.res, default None
        IncludeBoundary    : bool
            include the boundaries for the refinment, default True
        """
        self.Nfeild += 1
        if not np.iterable(ent_ID):
            ent_ID = [ent_ID]

        if dim == 2:
            typelist = "SurfacesList"
        else:
            typelist = "VolumesList"

        if res_out is None:
            res_out = self.res

        self.model.mesh.field.add("Constant", self.Nfeild)
        self.model.mesh.field.setNumbers(self.Nfeild, typelist, ent_ID)
        self.model.mesh.field.setNumber(self.Nfeild, "IncludeBoundary", IncludeBoundary)
        self.model.mesh.field.setNumber(self.Nfeild, "VIn", res_in)
        self.model.mesh.field.setNumber(self.Nfeild, "VOut", res_out)

        return self.Nfeild

    def refine_threshold(
        self, ent_ID, dim, dist_min, dist_max, res_min, res_max=None, N_samples=100
    ):
        """
        refine mesh resolution in a list of faces or volumes IDs

        Parameters
        ----------
        ent_ID    : list[int] or int
            ID or list of ID of the entities where the resolution should be changed
        res_in    : float
            resolution inside the entities
        dim    : int (2 or 3)
            dimention of the considerated entities
        res_out    : float
            resolution outside the entities if None take self.res, default None
        IncludeBoundary    : bool
            include the boundaries for the refinment, default True
        """
        self.Nfeild += 1
        if not np.iterable(ent_ID):
            ent_ID = [ent_ID]

        if dim == 2:
            typelist = "SurfacesList"
        elif dim == 1:
            typelist = "CurvesList"
        else:
            typelist = "PointsList"

        if res_max is None:
            res_max = self.res

        self.Nfeild += 1
        i_Dist_field = self.Nfeild
        self.model.mesh.field.add("Distance", i_Dist_field)
        self.model.mesh.field.setNumbers(i_Dist_field, typelist, ent_ID)
        self.model.mesh.field.setNumber(i_Dist_field, "Sampling", N_samples)

        self.Nfeild += 1
        self.model.mesh.field.add("Threshold", self.Nfeild)
        self.model.mesh.field.setNumber(self.Nfeild, "InField", i_Dist_field)
        self.model.mesh.field.setNumber(self.Nfeild, "SizeMin", res_min)
        self.model.mesh.field.setNumber(self.Nfeild, "SizeMax", res_max)
        self.model.mesh.field.setNumber(self.Nfeild, "DistMin", dist_min)
        self.model.mesh.field.setNumber(self.Nfeild, "DistMax", dist_max)
        self.model.mesh.field.setNumber(self.Nfeild, "Sigmoid", 1)

        return self.Nfeild

    def refine_min(self, feild_IDs):
        """
        refine mesh resolution taking the minimum value for a list of refinment fields

        Parameters
        ----------
        feild_IDs    : list[int]
            list of field from wich the minimum should be taken
        """
        self.Nfeild += 1

        self.model.mesh.field.add("Min", self.Nfeild)
        self.model.mesh.field.setNumbers(self.Nfeild, "FieldsList", feild_IDs)

        return self.Nfeild

    def refinement_callback(self, meshSizeCallback):
        """
        Add a call back function which is apply to the mesh size defined by fields and return
        the final mesh size

        Parameters
        ----------
        meshSizeCallback    : nrv.utils.MeshCallBack
        """
        self.model.mesh.setSizeCallback(meshSizeCallback)

    ##############################################################################################
    ################################   generate and saving methods  ##############################
    ##############################################################################################
    def generate(self):
        """
        genetate the mesh
        """
        if not self.is_generated:
            if self.Nfeild > 0:
                self.model.mesh.field.setAsBackgroundMesh(self.Nfeild)
            else:
                gmsh.option.setNumber("Mesh.CharacteristicLengthMin", 0.1 * self.res)
                gmsh.option.setNumber("Mesh.CharacteristicLengthMax", self.res)

            self.model.occ.synchronize()

            # print("nt:", gmsh.option.getNumber("General.NumThreads"))
            self.model.mesh.generate(self.D)
            self.is_generated = True

    def save(self, fname, generate=True):
        """
        Save mesh in fname in ".msh"

        Parameters
        ----------
        fname    : str
            path and name of saving file. If ends with ".msh" only save in ".msh" file
        """
        if generate == True:
            self.generate()
        fname = rmv_ext(fname)
        gmsh.write(fname + ".msh")
        self.file = fname

    def load(self, fname):
        """
        load mesh from ".msh" file
        """
        if ".msh" not in fname:
            fname += ".msh"
        if not os.path.isfile(fname):
            rise_warning(fname + " not found: Mesh could not be loaded")
        else:
            gmsh.merge(fname)
            self.file = fname

    def visualize(self, fname=None):
        if fname is None:
            self.generate()
            gmsh.fltk.run()
        elif fname is not None:
            self.save(fname=fname)
            os.system("gmsh " + self.file + ".msh")



from .log_interface import rise_warning


def set_attributes(my_object, attributes_dict):
    for key, value in attributes_dict.items():
        if key in my_object.__dict__:
            setattr(my_object, key, value)
        else:
            rise_warning(
                "trying to set a non existing attribute "
                + str(key)
                + " in "
                + str(type(my_object))
            )
    return 0



"""
NRV-electrodes
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler

import numpy as np

from ..backend.file_handler import json_load
from ..backend.log_interface import rise_error, rise_warning
from ..backend.NRV_Class import NRV_class

# enable faulthandler to ease "segmentation faults" debug
faulthandler.enable()


def is_FEM_electrode(elec):
    """
    Check if the electrode is a FEM based electrode
    """
    return issubclass(type(elec), FEM_electrode)


def is_CUFF_electrode(elec):
    """
    Check if the electrode is a LIFE electrode
    """
    return isinstance(elec, CUFF_electrode)


def is_LIFE_electrode(elec):
    """
    Check if the electrode is a LIFE electrode
    """
    return isinstance(elec, LIFE_electrode)


def is_analytical_electrode(elec):
    """
    Check if the electrode is an analytical based electrode
    """
    return not issubclass(type(elec), FEM_electrode)


def load_any_electrode(data):
    """
    return any kind of electrod context properties from a dictionary or a json file

    Parameters
    ----------
    data    : str or dict
        json file path or dictionary containing extracel_context information
    """
    if type(data) == str:
        elec_dic = json_load(data)
    else:
        elec_dic = data

    if elec_dic["type"] == "electrode":
        elec = electrode()
    elif elec_dic["type"] == "point source":
        elec = point_source_electrode(0, 0, 0)
    elif elec_dic["type"] == "FEM":
        elec = FEM_electrode("")
    elif elec_dic["type"] == "LIFE":
        elec = LIFE_electrode("", 0, 0, 0, 0, 0)
    elif elec_dic["type"] == "CUFF":
        elec = CUFF_electrode()
    elif elec_dic["type"] == "CUFF MP":
        elec = CUFF_MP_electrode()
    else:
        rise_error("Electrode type not recognizede")

    elec.load(elec_dic)
    return elec


def check_electrodes_overlap(elec1, elec2):
    """
    check if two FEM electrodes are overlaping

    Parameters
    ----------
    elec1:      FEM_electrode
        first electrode
    elec2:      FEM_electrode
        second electrode

    Returns
    -------
    test:           bool
        True if elec1 and elec2 are overlaping
    """
    if not is_FEM_electrode(elec1) or not is_FEM_electrode(elec2):
        return False
    if is_CUFF_electrode(elec1) and is_CUFF_electrode(elec2):
        dist_e = abs(elec1.x - elec2.x)
        len_min = (elec1.contact_length + elec2.contact_length) / 2
        if dist_e < len_min:
            return True
        else:
            return False
    elif is_LIFE_electrode(elec1) and is_LIFE_electrode(elec2):
        dist_e_x = abs(elec1.x - elec2.x)
        len_min_x = (elec1.length + elec2.length) / 2
        dist_e_yz = (elec1.y - elec2.y) ** 2 + (elec1.z - elec2.z) ** 2
        len_min_yz = ((elec1.D + elec2.D) / 2) ** 2
        if dist_e_x < len_min_x and dist_e_yz < len_min_yz:
            return True
        else:
            return False
    return False


class electrode(NRV_class):
    """
    Objet for generic electrode description. Each electrode has an ID and a position.
    """

    def __init__(self, ID=0):
        """
        Instantiation of a generic electrode

        Parameters
        ----------
        ID  : int
            electrode identification number, set to 0 by default
        """
        super().__init__()
        self.ID = ID
        self.footprint = np.asarray([])
        self.type = "electrode"
        self.is_multipolar = False
        self.x = None
        self.y = None
        self.z = None

    def save_electrode(self, save=False, fname="electrode.json"):
        rise_warning("save_electrode is a deprecated method use save")
        self.save(save=save, fname=fname)

    def load_electrode(self, data="electrode.json"):
        rise_warning("load_electrode is a deprecated method use load")
        self.load(data=data)

    def get_ID_number(self):
        """
        get the ID of a electrode

        Returns
        -------
        ID : int
            identification number of the electrode
        """
        return self.ID

    def set_ID_number(self, ID):
        """
        set the identification number of an electrode

        Parameters
        ----------
        ID  : int
            electrode identification number
        """
        self.ID = ID

    def get_footptint(self):
        """
        get the footprint of a electrode

        Returns
        -------
        footprint : np.array
            identification number of the electrode
        """
        return self.footprint

    def set_footprint(self, footprint):
        """
        set the footprint of a electrode

        Parameters
        ----------
        footprint : np.array
            array contaning the electrod linear footprint (electrode response at 1mA)
            on several space points
        """
        self.footprint = np.array(footprint)

    def clear_footprint(self):
        """
        clear the footprint of a electrode
        """
        self.footprint = np.array([])

    def compute_field(self, I):
        """Compute the external field using the Point source approximation

        Parameters
        ----------
        I       : float
            current value, in uA

        Returns
        -------
        v_ext   : np.array
            external voltage field value at specified coordinates with the specified material in mV
        """
        # convert uA in mA
        I_mA = I * 1e-3
        v_ext = I_mA * self.footprint
        return v_ext

    def translate(self, x=None, y=None, z=None):
        """
        Move electrode by translation

        Parameters
        ----------
        x   : float
            x axis value for the translation in um
        y   : float
            y axis value for the translation in um
        z   : float
            z axis value for the translation in um
        """
        if x is not None:
            self.x += x
        if y is not None:
            self.y += y
        if z is not None:
            self.z += z
        self.clear_footprint()


class point_source_electrode(electrode):
    """
    Point source electrode. Inherite from electrode. The electrode is punctual and act as a\
    monopole.
    """

    def __init__(self, x=0, y=0, z=0, ID=0):
        """
        Instantiation of a Point source electrode

        Parameters
        ----------
        x   : float
            x position of the electrode, in um
        y   : float
            y position of the electrode, in um
        z   : float
            z position of the electrode, in um
        ID  : int
            electrode identification number, set to 0 by default
        """
        super().__init__(ID)
        self.x = x
        self.y = y
        self.z = z

        self.type = "point source"

    def compute_footprint(self, x, y, z, mat):
        """
        Compute the linear footprint (electrode response at 1mA)

        Parameters
        ----------
        x       : float
            array like, x corrdinates where to compute the field, in um
        y       : float
            array like, y coordinates where to compute the field, in um
        z       : float
            array like, y coordinates where to compute the field, in um
        mat     : float
            material object, see Material help for more detail
        """
        if mat.is_isotropic():
            # 1e-6  on distances to stay in m (condivtivity specified in S/m
            self.footprint = 1.0 / (
                4
                * np.pi
                * np.sqrt(
                    (1e-6 * (self.x - x)) ** 2
                    + (1e-6 * (self.y - y)) ** 2
                    + (1e-6 * (self.z - z)) ** 2
                )
                * mat.sigma
            )
        else:
            sx = mat.sigma_yy * mat.sigma_zz
            sy = mat.sigma_xx * mat.sigma_zz
            sz = mat.sigma_xx * mat.sigma_yy
            # 1e-6  on distances to stay in m (condivtivity specified in S/m)
            self.footprint = 1.0 / (
                4
                * np.pi
                * np.sqrt(
                    sx * (1e-6 * (self.x - x)) ** 2
                    + sy * (1e-6 * (self.y - y)) ** 2
                    + sz * (1e-6 * (self.z - z)) ** 2
                )
            )


class FEM_electrode(electrode):
    """
    Electrode located in Finite Element Model in Comsol
    """

    def __init__(self, label, ID=0):
        """
        Instrantiation of a FEM electrode
        """
        super().__init__(ID)
        self.label = label
        self.ID = ID
        self.footprint = np.asarray([])
        self.is_volume = False
        self.type = "FEM"

    def set_footprint(self, V_1mA):
        """
        set the footprin of a FEM electrode

        Parameters:
        -----------
        V_1mA : list, array, numpy array
            Voltage response at 1mA
        """
        self.footprint = np.asarray(V_1mA)


class LIFE_electrode(FEM_electrode):
    """
    Longitudinal IntraFascicular Electrode for FEM models
    """

    def __init__(
        self,
        label="LIFE_1",
        D=25,
        length=1000,
        x_shift=0,
        y_c=0,
        z_c=0,
        ID=0,
    ):
        """
        Instantiation of a LIFE electrode

        Parameters
        ----------
        label   : str
            name of the electrode in the COMSOL file
        D       : float
            diameter of the electrode, in um
        length  : float
            length of the electrode, in um
        x_shift : float
            geometrical offset from the start (x=0) of the simulation
        y_c     : float
            y-coordinate of the center of the electrode, in um
        z_c     : float
            z-coordinate of the center of the electrode, in um
        """
        super().__init__(label, ID)
        self.D = D
        self.length = length
        self.x = x_shift
        self.y = y_c
        self.z = z_c
        self.type = "LIFE"

    def parameter_model(self, model, res="default"):
        """
        Parameter the model electrode with user specified dimensions

        Parameters
        ----------
        model : obj
            FEM COMSOL or Fenics simulation to parameter, se FEM or Extracellular for more details
        """
        if model.type == "COMSOL":
            model.set_parameter(self.label + "_D", str(self.D) + "[um]")
            model.set_parameter(self.label + "_Length", str(self.length) + "[um]")
            model.set_parameter(self.label + "_y_c", str(self.y) + "[um]")
            model.set_parameter(self.label + "_z_c", str(self.z) + "[um]")
            model.set_parameter(self.label + "_x_offset", str(self.x) + "[um]")
        else:
            model.add_electrode(
                elec_type=self.type,
                x_c=self.x + (self.length / 2),
                y_c=self.y,
                z_c=self.z,
                length=self.length,
                D=self.D,
                is_volume=self.is_volume,
                res=res,
            )


class CUFF_electrode(FEM_electrode):
    """
    CUFF electrode for FEM models
    """

    def __init__(
        self,
        label="",
        contact_length=100,
        is_volume=False,
        contact_thickness=None,
        insulator=True,
        insulator_length=None,
        insulator_thickness=None,
        x_center=0,
        insulator_offset=0,
        ID=0,
    ):
        """
        Instantiation of a LIFE electrode

        Parameters
        ----------
        label               : str
            name of the electrode in the COMSOL file
        x_center            :float
            x-position of the CUFF center in um, by default 0
            length of the CUFF electrod in um, by default 100
        contact_length      :float
            length along x of the contact site in um, by default 100
        is_volume   : bool
            if True the contact is kept on the mesh as a volume, by default True
        contact_thickness   :float
            thickness of the contact site in um, by default 5
        insulator            :bool
            remove insulator ring from the mesh (no conductivity), by default True
        insulator_thickness :float
            thickness of the insulator ring in um, by default 20
        insulator_length    :float
            length along x of the insulator ring in um, by default 1000
        """
        super().__init__(label, ID)
        self.contact_length = contact_length
        self.contact_thickness = contact_thickness
        self.x = x_center
        self.is_volume = is_volume
        self.insulator = insulator
        self.insulator_length = insulator_length
        self.insulator_thickness = insulator_thickness
        self.insulator_offset = insulator_offset
        self.type = "CUFF"

    def translate(self, x=None, y=None, z=None):
        """
        Move electrode by translation

        Parameters
        ----------
        x   : float
            x axis value for the translation in um
        y   : float
            y axis value for the translation in um
        z   : float
            z axis value for the translation in um
        """
        super().translate(x)

    def parameter_model(self, model, res="default"):
        """
        Parameter the model electrode with user specified dimensions

        Parameters
        ----------
        model : obj
            FEM COMSOL or Fenics simulation to parameter, se FEM or Extracellular for more details
        """
        if model.type == "COMSOL":
            model.set_parameter(
                self.label + "_contact_length", str(self.contact_length) + "[um]"
            )
            model.set_parameter(self.label + "_x_center", str(self.x) + "[um]")
            if self.contact_thickness is not None:
                model.set_parameter(
                    self.label + "_contact_thickness",
                    str(self.contact_thickness) + "[um]",
                )
            if self.insulator_thickness is not None:
                model.set_parameter(
                    self.label + "_insulator_thickness",
                    str(self.insulator_thickness) + "[um]",
                )
            if self.insulator_length is not None:
                model.set_parameter(
                    self.label + "_insulator_length",
                    str(self.insulator_length) + "[um]",
                )
        else:
            model.add_electrode(
                elec_type=self.type,
                insulator_length=self.insulator_length,
                is_volume=self.is_volume,
                insulator=self.insulator,
                insulator_thickness=self.insulator_thickness,
                contact_length=self.contact_length,
                contact_thickness=self.contact_thickness,
                insulator_offset=self.insulator_offset,
                x_c=self.x,
                res=res,
            )


class CUFF_MP_electrode(CUFF_electrode):
    """
    MultiPolar CUFF electrode for FEM models
    """

    def __init__(
        self,
        label="CUFF_MP",
        N_contact=4,
        contact_width=None,
        contact_length=100,
        is_volume=False,
        contact_thickness=None,
        insulator=True,
        insulator_length=None,
        insulator_thickness=None,
        x_center=0,
        insulator_offset=0,
        ID=0,
    ):
        """
        Instantiation of a LIFE electrode

        Parameters
        ----------
        label               : str
            name of the electrode in the COMSOL file
        x_center            :float
            x-position of the CUFF center in um, by default 0
        N_contact           :int
            Number of contact site of the electrode, by default 4
            length of the CUFF electrod in um, by default 100
        contact_length      :float
            length along x of the contact site in um, by default 100
        is_volume   : bool
            if True the contact is kept on the mesh as a volume, by default True
        contact_thickness   :float
            thickness of the contact site in um, by default 5
        inactive            :bool
            remove insulator ring from the mesh (no conductivity), by default True
        insulator_thickness :float
            thickness of the insulator ring in um, by default 20
        insulator_length    :float
            length along x of the insulator ring in um, by default 1000
        """
        super().__init__(
            label=label,
            contact_length=contact_length,
            is_volume=is_volume,
            contact_thickness=contact_thickness,
            insulator_length=insulator_length,
            insulator=insulator,
            insulator_thickness=insulator_thickness,
            x_center=x_center,
            insulator_offset=insulator_offset,
            ID=ID,
        )
        self.N_contact = N_contact
        self.contact_width = contact_width
        self.type = "CUFF MP"
        self.is_multipolar = True

    def parameter_model(self, model, res="default"):
        """
        Parameter the model electrode with user specified dimensions

        Parameters
        ----------
        model : obj
            FEM COMSOL or Fenics simulation to parameter, se FEM or Extracellular for more details
        """
        if model.type == "COMSOL":
            rise_warning("Multipolar CUFF not implemented on comsol")
        else:
            model.add_electrode(
                elec_type=self.type,
                N=self.N_contact,
                is_volume=self.is_volume,
                contact_width=self.contact_width,
                insulator_length=self.insulator_length,
                insulator_thickness=self.insulator_thickness,
                contact_length=self.contact_length,
                insulator=self.insulator,
                contact_thickness=self.contact_thickness,
                x_c=self.x,
                insulator_offset=self.insulator_offset,
                res=res,
            )



import os
import time

from dolfinx.fem import (
    Constant,
    Function,
    FunctionSpace,
    assemble_scalar,
    dirichletbc,
    form,
    locate_dofs_topological,
)
from dolfinx.fem.petsc import LinearProblem
from dolfinx.io.utils import XDMFFile
from mpi4py import MPI
from petsc4py.PETSc import ScalarType, Viewer
from ufl import (
    FiniteElement,
    Measure,
    MixedElement,
    TestFunction,
    TrialFunction,
    avg,
    inner,
    nabla_grad,
)

from ....backend.log_interface import pass_info, rise_error, rise_warning
from ....backend.file_handler import rmv_ext
from ....utils.units import S, V, m
from .fenics_materials import load_fenics_material
from .SimParameters import SimParameters
from .SimResult import read_gmsh, SimResult

# Lists of available solvers and conditioners
# go to https://petsc4py.readthedocs.io/en/stable/manual/ksp/ for more info

ksp_type_list = [
    "cg",
    "pipecg",
    "chebyshev",
    "groppcg",
    "pipecgrr",
    "cgne",
    "fcg",
    "cgls",
    "pipefcg",
    "nash",
    "stcg",
    "stcg",
    "qcg",
    "bicg",
    "bcgs",
    "ibcgs",
    "fbcgs",
    "fbcgsr",
    "symmlq",
    "bcgsl",
    "minres",
    "gmres",
    "fgmres",
    "dgmres",
    "pgmres",
    "pipefgmres",
    "lgmres",
    "cr",
    "gcr",
    "pipecr",
    "pipegcr",
    "fetidp",
    "cgs",
    "tfqmr",
    "tcqmr",
    "lsqr",
    "tsirm",
    "python",
    "preonly",
]

pc_type_list = [
    "lu",
    "ilu",
    "gamg",
    "jacobi",
    "sor",
    "eisenstat",
    "icc",
    "asm",
    "gasm",
    "bddc",
    "ksp",
    "composite",
    "cholesky",
    "none",
    "shell",
    "hypre",
]


class FEMSimulation(SimParameters):
    """
    Class usefull to solve the Static/Quasi-Static electrical current problem using FEM with
    FEniCSx algorithms (https://fenicsproject.org).

    / add equation /

    The problem parameters (domains and boundaries condition) can be define using SimParameters methods
    Contains methods to prepare the matrix sytstem, to solve it and to access the results.

    Inherit from SimParameters class. see SimParameters for further detail
    """

    def __init__(
        self,
        D=3,
        mesh_file="",
        mesh=None,
        data=None,
        elem=None,
        ummesh=True,
        comm=MPI.COMM_SELF,
        rank=0,
    ):
        """
        initialisation of the FEMSimulation:

        Parameters
        ----------
        D               : int
            dim of the mesh, by default 3
            NB: only 3 is implemented
        mesh_file       : str
            mesh directory and file name: by default ""
        mesh            : None or MshCreator
            if not None, MshCreator from which the mesh sould be used, by default None
            (see MshCreator and NerveMshCreator for more details)
        data            : str, dict or SimParameters
            if not None, load SimParameters attribute from data, by default None
            (see SimParameters.load)
        elem            :tupple (str, int)
            if None, ("Lagrange", 1), else (element type, element order), by default None
        ummesh          : bool
            if True the scale of mesh space dimensions should be (um), else (m), by default True
            Usefull to link the update materials conductivity as in NRV conductivities are in S/m
            but NerveMshCreator space scale is um)
        comm            : int
            The MPI communicator to use for mesh creation, by default MPI.COMM_SELF
        rank            : int
            The rank the Gmsh model is initialized on, by default 0
        """
        super().__init__(D=D, mesh_file=mesh_file, data=data)
        if elem is not None:
            self.elem = elem
        else:
            self.elem = ("Lagrange", 1)
        self.multi_elem = None

        # Mesh and Meshtag
        self.mesh = mesh

        if self.mesh is None:
            self.mesh = self.mesh_file
        self.domain = None
        self.subdomains = None
        self.boundaries = None

        # Space and mesure
        self.V = None
        self.V_DG = None
        self.dS = None
        self.ds = None
        self.dx = None
        if ummesh:
            self.mat_unit = "S/um"
        else:
            self.mat_unit = "S/m"

        # Multimesh parmeters
        self.Nspace = 1

        # Bilinear and linear forms
        self.a = None
        self.L = None

        self.mat_list_ID = []
        self.sigma_list = []
        self.mat_list = []
        self.mat_list_space = []
        self.mixedvouts = []

        # gather fen_mat name of all domain and internal layer
        self.mat_map = {}
        self.Neuman_list = {}

        # Solver parameters
        self.petsc_opt = {
            "ksp_type": "cg",
            "pc_type": "ilu",
            "ksp_rtol": 1e-4,
            "ksp_atol": 1e-7,
            "ksp_max_it": 1000,
        }
        self.cg_problem = None
        self.dg_problem = None
        self.u = None
        self.mixedvout = None
        self.vout = None
        self.result = None

        # Mcore attributes
        self.comm = comm
        self.rank = rank

        # Timmers
        self.solving_timer = 0

        self.bcs = []
        # added for overzriting false option
        self.file = []

        self.data_status = True
        self.domain_status = False
        self.dirichlet_BC_status = False
        self.neumann_BC_status = False
        self.bilinear_form_status = False
        self.jump_status = False
        self.linear_form_status = False
        self.prepared_status = False
        self.solve_status = False
        self.to_merge = True

    #####################################################
    ########### SimParameter methods overload ###########
    #####################################################

    def add_inboundary(
        self,
        mesh_domain,
        in_domains,
        thickness,
        mat_pty=None,
        mat_file=None,
        mat_perm=None,
        ID=None,
    ):
        super().add_inboundary(
            mesh_domain,
            in_domains,
            thickness,
            mat_pty,
            mat_file,
            mat_perm,
            ID,
        )
        if self.prepared_status:
            if mesh_domain in self.mat_map:
                mat = load_fenics_material(mat_pty)
                self.mat_map[mesh_domain].load_from_mat(mat)
            else:
                rise_warning(
                    "Domain not added: new domain cannot be added between 2 simulations,\
                             (set domain before simulation or create a new one)"
                )

    def add_domain(
        self, mesh_domain, mat_pty=None, mat_file=None, mat_perm=None, ID=None
    ):
        super().add_domain(mesh_domain, mat_pty, mat_file, mat_perm, ID)
        if self.prepared_status:
            if mesh_domain in self.mat_map:
                mat = load_fenics_material(mat_pty)
                self.mat_map[mesh_domain].load_from_mat(mat)
            else:
                rise_warning(
                    "Domain not added: new domain cannot be added between 2 simulations,\
                             (set domain before simulation or create a new one)"
                )

    #####################################################
    ############ Prepare the matrix sytstem #############
    #####################################################

    def prepare_sim(self, **kwargs):
        """
        Prepare Bilinear form, Linear form and boundary conditions using paramters and kwargs
        to set the variable defined Neumann boundary conditions (see SimParameters.add_boundary)
        If FEMSimulation already defined, can be used to modify variable defined NBC
        """
        t0 = time.time()
        if self.data_status:
            self.args = kwargs
            pass_info("Static/Quasi-Static electrical current problem")
            if self.D == 1:
                rise_error("1D not implemented yet")
            else:
                # Initialize the domain
                if not self.domain_status:
                    self.__init_domain()
                # DIRICHLET BOUNDARY CONDITIONS
                if not self.dirichlet_BC_status:
                    self.__set_dirichlet_BC()
                # DEFINING THE VARIATIONAL PROBLEM
                self.mixedvout = TrialFunction(self.V)
                self.u = TestFunction(self.V)
                # defining the bilinear form a(vout,u)
                if not self.bilinear_form_status:
                    self.__set_bilinear_form()
                # defining the linear form L(u)
                # NB1: in the case of the current-problem, the source term is nul
                if not self.linear_form_status:
                    self.__set_linear_form()
                # NEUMANN BOUNDARY CONDITIONS
                self.__set_neumann_BC()
            self.prepared_status = True
            self.solving_timer += time.time() - t0
        else:
            rise_warning("no parameters are set, Sim can not be prepared")

    def __init_domain(self):
        """
        internal use only:
        """
        if not self.domain_status:
            # RECOVERING THE GEOMETRY
            self.domain, self.subdomains, self.boundaries = read_gmsh(
                self.mesh, comm=self.comm, rank=self.rank, gdim=3
            )
            # SPACE FOR INTEGRATION
            if self.inbound:
                self.Nspace = self.Ninboundaries + 1
                ME = [
                    FiniteElement(self.elem[0], self.domain.ufl_cell(), self.elem[1])
                    for k in range(self.Nspace)
                ]
                self.multi_elem = MixedElement(ME)
            else:
                self.multi_elem = self.elem
            self.V = FunctionSpace(self.domain, self.multi_elem)
            # MEASURES FOR INTEGRATION
            self.dx = Measure("dx", domain=self.domain, subdomain_data=self.subdomains)
            self.ds = Measure("ds", domain=self.domain, subdomain_data=self.boundaries)
            self.dS = Measure("dS", domain=self.domain, subdomain_data=self.boundaries)
            self.domain_status = True

    def __set_dirichlet_BC(self):
        """
        internal use only: set the Dirichlet boundary condition from parameters
        NB: DBC cannot be change between simulations
        """
        self.bcs = []
        for i in self.boundaries_list:
            bound = self.boundaries_list[i]
            condition = bound["condition"]
            if condition.lower() == "dirichlet":
                value = Constant(self.domain, ScalarType(float(bound["value"])))
                label = self.boundaries.find(int(bound["mesh_domain"]))
                if not self.inbound:
                    dofs = locate_dofs_topological(
                        self.V, self.domain.topology.dim - 1, label
                    )
                    self.bcs.append(dirichletbc(value, dofs, self.V))
                else:
                    i_space = self.get_space_of_domain(bound["mesh_domain_3D"])
                    dofs = locate_dofs_topological(
                        self.V.sub(i_space), self.domain.topology.dim - 1, label
                    )
                    self.bcs.append(dirichletbc(value, dofs, self.V.sub(i_space)))
        if not self.bcs:
            rise_warning(
                "no Dirichlet Condition implemented on the Computed Field, Simulation maybe unsuccessful"
            )
        self.dirichlet_BC_status = True

    def __set_neumann_BC(self):
        """
        internal use only: set the Neuman boundary condition from parameters or update
        variable between to simulation
        NB: Only variable NBC (see SimParameters.add_boundary) can be changed between simulations
        """
        if not self.neumann_BC_status:
            for i_bound in self.boundaries_list:
                bound = self.boundaries_list[i_bound]
                condition = bound["condition"]
                if condition.lower() in "neumann":
                    dom = int(bound["mesh_domain"])
                    if "value" in bound:
                        self.Neuman_list[i_bound] = Constant(
                            self.domain, ScalarType(bound["value"])
                        )
                    elif "variable" in bound:
                        self.Neuman_list[i_bound] = Constant(
                            self.domain, ScalarType(self.args[bound["variable"]])
                        )
                    else:
                        rise_error(
                            "A Neuman Boundary condition must be associated with a value or variable"
                        )
                    if not self.inbound:
                        self.L = self.L + self.Neuman_list[i_bound] * self.u * self.ds(
                            dom
                        )
                    else:
                        i_space = self.get_space_of_domain(bound["mesh_domain_3D"])
                        self.L = self.L + self.Neuman_list[i_bound] * self.u[
                            i_space
                        ] * self.ds(dom)
            self.neumann_BC_status = True
        else:
            for i_bound in self.boundaries_list:
                bound = self.boundaries_list[i_bound]
                condition = bound["condition"]
                if condition.lower() in "neumann" and "variable" in bound:
                    if bound["variable"] in self.args:
                        self.Neuman_list[i_bound].value = self.args[bound["variable"]]

    def __set_bilinear_form(self):
        """
        internal use only: set the bilinear form a(vout, u) from the parameters
        """
        pass_info("FEN4NRV: preparing the bilinear form")
        self.__set_material_map()
        for i_space in range(self.Nspace):
            for i_domain in self.domains_list:
                dom = self.get_mixedspace_domain(i_space=i_space, i_domain=i_domain)
                if self.a is None:
                    self.a = self.__get_static_component(i_domain, dom, i_space)
                else:
                    self.a += self.__get_static_component(i_domain, dom, i_space)
        if self.inbound:
            self.__set_jump()
        self.bilinear_form_status = True

    def __get_static_component(self, i_dom, i_mat, i_space):
        """
        Set a static componnent of the bimlinear form:
        a = ∇v[i_space] x 𝜎[i_mat]∇u[i_space] dx(i_dom)
        Parameters
        ----------
        i_dom    : int
            id of the domain on which the component should be set
        i_mat       : int
            id of the material corresponding domain on which the component should be set
        i_space     : int
            id of the i_space
        """
        if not self.inbound:
            return inner(
                nabla_grad(self.mixedvout),
                self.mat_map[i_mat].sigma_fen * nabla_grad(self.u),
            ) * self.dx(i_dom)
        else:
            return inner(
                nabla_grad(self.mixedvout[i_space]),
                self.mat_map[i_mat].sigma_fen * nabla_grad(self.u[i_space]),
            ) * self.dx(i_dom)

    def __set_jump(self):
        """
        internal use only: set the jump for all internale boundary to mimic the thin layers
        """
        for i_ibound in self.inboundaries_list:
            in_space, out_space = self.get_spaces_of_ibound(i_ibound)
            local_thickness = Constant(
                self.domain, ScalarType(self.inboundaries_list[i_ibound]["thickness"])
            )
            jmp_v = avg(self.mixedvout[out_space]) - avg(self.mixedvout[in_space])
            jmp_u = avg(self.u[out_space]) - avg(self.u[in_space])
            self.a += (
                self.mat_map[i_ibound].sigma_fen
                / local_thickness
                * jmp_u
                * jmp_v
                * self.dS(i_ibound)
            )
        self.jump_status = True

    def __set_material_map(self):
        """
        internal use only: build a dictionnary mat_map containing a material for every domain and layer
        """
        if self.mat_unit == "S/um":
            UN = S / m
        else:
            UN = 1
        for dom, pty in self.mat_pty_map.items():
            self.mat_map[dom] = load_fenics_material(pty)
            self.mat_map[dom].update_fenics_sigma(
                domain=self.domain, elem=self.elem, UN=UN, id=dom
            )

    def __set_linear_form(self):
        """
        internal use only: set the linear form L(u) from the parameters
        """
        pass_info("FEN4NRV: preparing the linear form")
        # Check if quicker without
        c_0 = Constant(self.domain, ScalarType(0.0))
        if not self.inbound:
            self.L = c_0 * self.u * self.dx
        else:
            for i_space in range(self.Nspace):
                self.L = c_0 * self.u[i_space] * self.dx
        self.linear_form_status = True

    #####################################################
    ################# Solve the sytstem #################
    #####################################################

    def get_solver_opt(self):
        """
        get krylov solver options
        """
        return self.petsc_opt

    def set_solver_opt(
        self, ksp_type=None, pc_type=None, ksp_rtol=None, ksp_atol=None, ksp_max_it=None
    ):
        """
        set krylov solver options

        Parameters
        ----------
        ksp_type        : str
            value to set for ksp_type (solver type), if None don"t change, None by default
        pc_type         : str
            value to set for pc_type (preconditioner type), if None don"t change, None by default
            list of possible
        ksp_rtol        : float
            value to set for ksp_rtol (relative tolerance), if None don"t change, None by default
        ksp_atol        : float
            value to set for ksp_atol (absolute tolerance), if None don"t change, None by default
        ksp_max_it      : int
            value to set for ksp_max_it (max number of iterations), if None don"t change, None by default
        """
        if ksp_type is not None:
            if ksp_type in ksp_type_list:
                self.petsc_opt["ksp_type"] = ksp_type
            else:
                rise_warning(ksp_type + " not set, should be in:\n" + ksp_type_list)
        if pc_type is not None:
            if pc_type in pc_type_list:
                self.petsc_opt["pc_type"] = pc_type
            else:
                rise_warning(pc_type + " not set, should be in:\n" + pc_type_list)
        if ksp_rtol is not None:
            self.petsc_opt["ksp_rtol"] = ksp_rtol
        if ksp_atol is not None:
            self.petsc_opt["ksp_atol"] = ksp_atol
        if ksp_max_it is not None:
            self.petsc_opt["ksp_max_it"] = ksp_max_it

    def set_result_merging(self, to_merge=None):
        if isinstance(to_merge, bool):
            self.to_merge = to_merge
        else:
            self.to_merge = not self.to_merge

    def solve(self, overwrite=False):
        """
        Assemble and solve the linear problem

        Parameters
        ----------
        overwrite   : bool
            if true modify the existing sim_res value, else create a new one. by default False
        Returns
        -------
        self.results    : SimResult
            SimResult containing the result of the resulting field of the FEM simulation
        """
        t0 = time.time()
        pass_info("FEN4NRV: solving electrical potential")
        if self.cg_problem is None:
            self.cg_problem = LinearProblem(
                self.a, self.L, bcs=self.bcs, petsc_options=self.petsc_opt
            )
        self.mixedvout = self.cg_problem.solve()
        self.solve_status = True

        if self.inbound and self.to_merge:
            V_sol = self.__merge_mixed_solutions()
        else:
            self.vout = self.mixedvout.copy()
            V_sol = self.V

        # return simulation result
        self.result = SimResult()
        if not overwrite:
            vout = Function(V_sol)
            vout.x.array[:] = self.vout.x.array[:]
        else:
            vout = self.vout
        self.result.set_sim_result(
            mesh_file=self.mesh_file,
            domain=self.domain,
            elem=self.multi_elem,
            V=V_sol,
            vout=vout,
            comm=self.domain.comm,
        )
        self.solving_timer += time.time() - t0
        pass_info("FEN4NRV: solved in " + str(self.solving_timer) + " s")
        return self.result

    def __merge_mixed_solutions(self):
        if self.dg_problem is None:
            self.mixedvouts = self.mixedvout.split()
            self.V_DG = FunctionSpace(
                self.domain, ("Discontinuous Lagrange", self.elem[1])
            )
            u, v = TrialFunction(self.V_DG), TestFunction(self.V_DG)
            adg = u * v * self.dx
            Ldg = 0
            for i_domain in self.domainsID:
                i_space = self.get_space_of_domain(i_domain)
                Ldg += v * self.mixedvout[i_space] * self.dx(i_domain)
            self.dg_problem = LinearProblem(
                adg, Ldg, bcs=[], petsc_options=self.petsc_opt
            )
        else:
            mixedvouts = self.mixedvout.split()
            for i in range(len(mixedvouts)):
                self.mixedvouts[i].x.array[:] = mixedvouts[i].x.array
        self.vout = self.dg_problem.solve()
        return self.V_DG

    #####################################################
    ################ Access the results #################
    #####################################################

    def solver_info(self, txt_fname="solver.txt"):
        solver = self.cg_problem.solver
        viewer = Viewer().createASCII(txt_fname)
        solver.view(viewer)
        solver_output = open(txt_fname, "r")
        for line in solver_output.readlines():
            print(line)

    def solve_and_save_sim(self, filename, save=True):
        if not self.solve_status:
            self.solve()

        fname = rmv_ext(filename)
        if not (fname == filename or fname + ".xdmf" == filename):
            rise_warning("Extension of solution will be save in xdmf files")
        with XDMFFile(self.domain.comm, fname + ".xdmf", "w") as file:
            file.write_mesh(self.domain)
            file.write_function(self.vout)
        return self.result

    def get_timers(self):
        return self.solving_timer

    def visualize_mesh(self):
        os.system("gmsh " + self.mesh_file + ".msh")

    def get_domain_potential(self, dom_id, dim=2, space=0):
        if dim == 2:
            do = self.ds
        elif dim == 3:
            do = self.dx
        Surf = assemble_scalar(form(1 * do(dom_id)))
        if self.to_merge:
            return assemble_scalar(form(self.vout * do(dom_id))) / Surf * V
        else:
            return assemble_scalar(form(self.vout[space] * do(dom_id))) / Surf * V



"""
NRV-compileMods
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""

import os

# test if neuron correctly installed
# warning: difficult as neuron can be 'mocked' on github and rtd for delploy and help
try:
    import neuron as nrn  # this still work with fake neuron (mock)

    nrn_version = nrn.__version__  # this fails :)
except AttributeError:
    from .log_interface import rise_warning

    # no Neuron or fake neuron
    rise_warning(
        "No mods compiled as the environnement variable NRVPATH does not exist"
    )
else:
    # if true neuron installation
    dir_path = os.environ["NRVPATH"] + "/_misc/mods/"
    ls = os.listdir(dir_path)
    test_mods = False
    i = 0
    while (not test_mods) and i < len(ls):
        test_mods = "." not in ls[i]
        i += 1

    def NeuronCompile():
        # path2compiled_mods =  dir_path + "/mods/x86_64"
        path2_mods = dir_path
        os.system("chmod +x " + os.environ["NRVPATH"] + "/nrv2calm")
        os.system("cd " + path2_mods + "&& nrnivmodl")

    if not test_mods:
        print("Mods files are not compiled, executing nrnivmodl...")
        NeuronCompile()
        print("Compilation done")



"""
Access and modify NRV Parameters
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""

# from .log_interface import rise_warning


class NRV_singleton(type):
    """
    Should be used as metaclass to define singleton classes

    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(NRV_singleton, cls).__call__(*args, **kwargs)
        else:
            print("WARNING : Class", cls, "is a sigleton cannot be re-instantiated")
            # rise_warning('Class', cls, 'is a sigleton cannot be re-instantiated')
        return cls._instances[cls]



"""
Access and modify NRV Parameters
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import configparser
import os

from .NRV_Singleton import NRV_singleton


class nrv_parameters(metaclass=NRV_singleton):
    """
    A class for NRV parameters used to gather parameters
    """

    def __init__(self):
        """
        Initialize the class for parameters
        """
        self.dir_path = os.environ["NRVPATH"] + "/_misc"
        self.config_fname = self.dir_path + "/NRV.ini"

        self.load()

    def save(self):
        """
        Saving the parameters
        """
        with open(self.config_fname, "w") as configfile:
            self.machine_config.write(configfile)

    def load(self, fname=None):
        """
        Loading the parameters
        """
        if fname is None:
            fname = self.config_fname

        self.machine_config = configparser.ConfigParser()
        self.machine_config.read(self.config_fname)

        # GMSH
        self.GMSH_Ncores = int(self.machine_config.get("GMSH", "GMSH_CPU"))
        self.GMSH_Status = self.machine_config.get("GMSH", "GMSH_STATUS") == "True"
        # LOG
        self.LOG_Status = self.machine_config.get("LOG", "LOG_STATUS") == "True"
        self.VERBOSITY_LEVEL = int(self.machine_config.get("LOG", "VERBOSITY_LEVEL"))

    def get_nrv_verbosity(self):
        """
        get general verbosity level
        """
        return self.VERBOSITY_LEVEL

    def set_nrv_verbosity(self, i):
        """
        set general verbosity level
        Parameters
        ----------
        O: None
        1: + Error
        2: + Warning
        3: + Info
        4: + Debug
        NB: to add Debug verbosity to the log VERBOSITY_LEVEL has to be set to 4
        """
        self.VERBOSITY_LEVEL = i

    def get_gmsh_ncore(self):
        """
        get gmsh core number
        """
        return self.GMSH_Ncores

    def set_gmsh_ncore(self, n):
        """
        set gmsh core number
        """
        self.GMSH_Ncores = n


###########################
#   Parameter singleton   #
###########################

parameters = nrv_parameters()
